from http import HTTPStatus
from unittest.mock import patch

from django.contrib.admin.sites import AdminSite
from django.contrib.auth.models import User
from django.test import TestCase

from app_utils.testing import create_user_from_evecharacter

from freight.admin import ContractAdmin
from freight.models import Contract

from ..tests.testdata.factories import (
    create_contract,
    create_contract_handler,
    create_pricing,
)
from .testdata.helpers import create_entities_from_characters, create_locations

MODULE_PATH = "freight.admin"


class MockRequest(object):
    def __init__(self, user=None):
        self.user = user


class TestContractAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        create_locations()
        create_entities_from_characters()
        cls.modeladmin = ContractAdmin(model=Contract, admin_site=AdminSite())
        cls.user = User.objects.create_superuser("Clark Kent")
        _, character_ownership = create_user_from_evecharacter(90000001)
        cls.handler = create_contract_handler(character=character_ownership)
        cls.contract = create_contract(handler=cls.handler)

    @patch(MODULE_PATH + ".ContractAdmin.message_user", spec=True)
    @patch(MODULE_PATH + ".Contract.send_pilot_notification")
    def test_should_send_pilots_notification(
        self, mock_send_pilot_notification, mock_message_user
    ):
        # given
        obj_qs = Contract.objects.filter(pk=self.contract.pk)
        # when
        self.modeladmin.send_pilots_notification(MockRequest(self.user), obj_qs)
        # then
        self.assertEqual(mock_send_pilot_notification.call_count, 1)
        self.assertTrue(mock_message_user.called)

    @patch(MODULE_PATH + ".ContractAdmin.message_user", spec=True)
    @patch(MODULE_PATH + ".Contract.send_customer_notification")
    def test_should_send_customer_notification(
        self, mock_send_customer_notification, mock_message_user
    ):
        # given
        obj_qs = Contract.objects.filter(pk=self.contract.pk)
        # when
        self.modeladmin.send_customer_notification(MockRequest(self.user), obj_qs)
        # then
        self.assertEqual(mock_send_customer_notification.call_count, 1)
        self.assertTrue(mock_message_user.called)

    def test_should_open_list(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/admin/freight/contract/")
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)


class TestPricingAdmin(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.user = User.objects.create_superuser("Clark Kent")
        create_locations()
        create_pricing(
            start_location_id=60003760,
            end_location_id=1022167642188,
            price_base=100000000,
        )

    def test_should_open_list(self):
        # given
        self.client.force_login(self.user)
        # when
        response = self.client.get("/admin/freight/pricing/")
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)



from unittest.mock import patch

from django.core.exceptions import ValidationError

from app_utils.testing import NoSocketsTestCase

from freight.models import ContractHandler, Location, Pricing

from ..testdata.factories import create_pricing
from ..testdata.helpers import create_contract_handler_w_contracts, create_locations

MODULE_PATH = "freight.models.routes"


class TestLocation(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.jita, cls.amamake, _ = create_locations()

    def test_str(self):
        self.assertEqual(
            str(self.jita.name), "Jita IV - Moon 4 - Caldari Navy Assembly Plant"
        )

    def test_repr(self):
        expected = (
            f"Location(pk={self.amamake.pk}, "
            "name='Amamake - 3 Time Nearly AT "
            "Winners')"
        )
        self.assertEqual(repr(self.amamake), expected)

    def test_category(self):
        self.assertEqual(self.jita.category, Location.Category.STATION_ID)

    def test_solar_system_name_station(self):
        self.assertEqual(self.jita.solar_system_name, "Jita")

    def test_solar_system_name_structure(self):
        self.assertEqual(self.amamake.solar_system_name, "Amamake")

    def test_location_name_station(self):
        self.assertEqual(self.jita.location_name, "Caldari Navy Assembly Plant")

    def test_location_name_structure(self):
        self.assertEqual(self.amamake.location_name, "3 Time Nearly AT Winners")


class TestPricing(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.handler, _ = create_contract_handler_w_contracts()
        cls.jita = Location.objects.get(id=60003760)
        cls.amamake = Location.objects.get(id=1022167642188)
        cls.amarr = Location.objects.get(id=60008494)

    @patch(MODULE_PATH + ".FREIGHT_FULL_ROUTE_NAMES", False)
    def test_str(self):
        p = Pricing(
            start_location=self.jita, end_location=self.amamake, price_base=50000000
        )
        expected = "Jita <-> Amamake"
        self.assertEqual(str(p), expected)

    def test_repr(self):
        p = Pricing(
            start_location=self.jita, end_location=self.amamake, price_base=50000000
        )
        expected = (
            f"Pricing(pk={p.pk}, "
            "name='Jita IV - Moon 4 - Caldari Navy Assembly Plant "
            "<-> Amamake - 3 Time Nearly AT Winners')"
        )
        self.assertEqual(repr(p), expected)

    @patch(MODULE_PATH + ".FREIGHT_FULL_ROUTE_NAMES", False)
    def test_name_from_settings_short(self):
        p = Pricing(
            start_location=self.jita, end_location=self.amamake, price_base=50000000
        )
        self.assertEqual(p.name, "Jita <-> Amamake")

    def test_name_short(self):
        p = Pricing(
            start_location=self.jita, end_location=self.amamake, price_base=50000000
        )
        self.assertEqual(p.name_short, "Jita <-> Amamake")

    @patch(MODULE_PATH + ".FREIGHT_FULL_ROUTE_NAMES", True)
    def test_name_from_settings_full(self):
        p = Pricing(
            start_location=self.jita, end_location=self.amamake, price_base=50000000
        )
        self.assertEqual(
            p.name,
            "Jita IV - Moon 4 - Caldari Navy Assembly Plant <-> "
            "Amamake - 3 Time Nearly AT Winners",
        )

    def test_name_full(self):
        p = Pricing(
            start_location=self.jita, end_location=self.amamake, price_base=50000000
        )
        self.assertEqual(
            p.name_full,
            "Jita IV - Moon 4 - Caldari Navy Assembly Plant <-> "
            "Amamake - 3 Time Nearly AT Winners",
        )

    def test_create_pricings(self):
        # first pricing
        create_pricing(
            start_location=self.jita,
            end_location=self.amamake,
            price_base=500000000,
        )
        # pricing with different route
        create_pricing(
            start_location=self.amarr,
            end_location=self.amamake,
            price_base=250000000,
        )
        # pricing with reverse route then pricing 1
        create_pricing(
            start_location=self.amamake,
            end_location=self.jita,
            price_base=350000000,
        )

    def test_create_pricing_no_2nd_bidirectional_allowed(self):
        create_pricing(
            start_location=self.jita,
            end_location=self.amamake,
            price_base=500000000,
            is_bidirectional=True,
        )
        p = create_pricing(
            start_location=self.amamake,
            end_location=self.jita,
            price_base=500000000,
            is_bidirectional=True,
        )
        with self.assertRaises(ValidationError):
            p.clean()

    def test_create_pricing_no_2nd_unidirectional_allowed(self):
        create_pricing(
            start_location=self.jita,
            end_location=self.amamake,
            price_base=500000000,
            is_bidirectional=True,
        )
        p = create_pricing(
            start_location=self.amamake,
            end_location=self.jita,
            price_base=500000000,
            is_bidirectional=False,
        )
        p.clean()
        # this test case has been temporary inverted to allow users
        # to migrate their pricings
        """
        with self.assertRaises(ValidationError):
            p.clean()
        """

    def test_create_pricing_2nd_must_be_unidirectional_a(self):
        create_pricing(
            start_location=self.jita,
            end_location=self.amamake,
            price_base=500000000,
            is_bidirectional=False,
        )
        p = create_pricing(
            start_location=self.amamake,
            end_location=self.jita,
            price_base=500000000,
            is_bidirectional=True,
        )
        with self.assertRaises(ValidationError):
            p.clean()

    def test_create_pricing_2nd_ok_when_unidirectional(self):
        create_pricing(
            start_location=self.jita,
            end_location=self.amamake,
            price_base=500000000,
            is_bidirectional=False,
        )
        p = create_pricing(
            start_location=self.amamake,
            end_location=self.jita,
            price_base=500000000,
            is_bidirectional=False,
        )
        p.clean()

    def test_name_uni_directional(self):
        p = Pricing(
            start_location=self.jita,
            end_location=self.amamake,
            price_base=50000000,
            is_bidirectional=False,
        )
        self.assertEqual(p.name, "Jita -> Amamake")

    def test_get_calculated_price(self):
        p = Pricing()
        p.price_per_volume = 50
        self.assertEqual(p.get_calculated_price(10, 0), 500)

        p = Pricing()
        p.price_per_collateral_percent = 2
        self.assertEqual(p.get_calculated_price(10, 1000), 20)

        p = Pricing()
        p.price_per_volume = 50
        p.price_per_collateral_percent = 2
        self.assertEqual(p.get_calculated_price(10, 1000), 520)

        p = Pricing()
        p.price_base = 20
        self.assertEqual(p.get_calculated_price(10, 1000), 20)

        p = Pricing()
        p.price_min = 1000
        self.assertEqual(p.get_calculated_price(10, 1000), 1000)

        p = Pricing()
        p.price_base = 20
        p.price_per_volume = 50
        self.assertEqual(p.get_calculated_price(10, 1000), 520)

        p = Pricing()
        p.price_base = 20
        p.price_per_volume = 50
        p.price_min = 1000
        self.assertEqual(p.get_calculated_price(10, 1000), 1000)

        p = Pricing()
        p.price_base = 20
        p.price_per_volume = 50
        p.price_per_collateral_percent = 2
        p.price_min = 500
        self.assertEqual(p.get_calculated_price(10, 1000), 540)

        with self.assertRaises(ValueError):
            p.get_calculated_price(-5, 0)

        with self.assertRaises(ValueError):
            p.get_calculated_price(50, -5)

        p = Pricing()
        p.price_base = 0
        self.assertEqual(p.get_calculated_price(None, None), 0)

        p = Pricing()
        p.price_per_volume = 50
        self.assertEqual(p.get_calculated_price(10, None), 500)

        p = Pricing()
        p.price_per_collateral_percent = 2
        self.assertEqual(p.get_calculated_price(None, 100), 2)

    def test_get_contract_pricing_errors(self):
        p = Pricing()
        p.price_base = 50
        self.assertIsNone(p.get_contract_price_check_issues(10, 20, 50))

        p = Pricing()
        p.price_base = 500
        p.volume_max = 300
        self.assertIsNotNone(p.get_contract_price_check_issues(350, 1000))

        p = Pricing()
        p.price_base = 500
        p.volume_min = 100
        self.assertIsNotNone(p.get_contract_price_check_issues(50, 1000))

        p = Pricing()
        p.price_base = 500
        p.collateral_max = 300
        self.assertIsNotNone(p.get_contract_price_check_issues(350, 1000))

        p = Pricing()
        p.price_base = 500
        p.collateral_min = 300
        self.assertIsNotNone(p.get_contract_price_check_issues(350, 200))

        p = Pricing()
        p.price_base = 500
        self.assertIsNotNone(p.get_contract_price_check_issues(350, 200, 400))

        p = Pricing()
        p.price_base = 500
        with self.assertRaises(ValueError):
            p.get_contract_price_check_issues(-5, 0)

        with self.assertRaises(ValueError):
            p.get_contract_price_check_issues(50, -5)

        with self.assertRaises(ValueError):
            p.get_contract_price_check_issues(50, 5, -5)

    def test_collateral_min_allows_zero(self):
        p = Pricing()
        p.price_base = 500
        p.collateral_min = 0
        self.assertIsNone(p.get_contract_price_check_issues(350, 0))

    def test_collateral_min_allows_none(self):
        p = Pricing()
        p.price_base = 500
        self.assertIsNone(p.get_contract_price_check_issues(350, 0))

    def test_zero_collateral_allowed_for_collateral_pricing(self):
        p = Pricing()
        p.collateral_min = 0
        p.price_base = 500
        p.price_per_collateral_percent = 2

        self.assertIsNone(p.get_contract_price_check_issues(350, 0))
        self.assertEqual(p.get_calculated_price(350, 0), 500)

    def test_requires_volume(self):
        self.assertTrue(Pricing(price_per_volume=10000).requires_volume())
        self.assertTrue(Pricing(volume_min=10000).requires_volume())
        self.assertTrue(
            Pricing(price_per_volume=10000, volume_min=10000).requires_volume()
        )
        self.assertFalse(Pricing().requires_volume())

    def test_requires_collateral(self):
        self.assertTrue(Pricing(price_per_collateral_percent=2).requires_collateral())
        self.assertTrue(Pricing(collateral_min=50000000).requires_collateral())
        self.assertTrue(
            Pricing(
                price_per_collateral_percent=2, collateral_min=50000000
            ).requires_collateral()
        )
        self.assertFalse(Pricing().requires_collateral())

    def test_clean_force_error(self):
        p = Pricing()
        with self.assertRaises(ValidationError):
            p.clean()

    def test_is_fix_price(self):
        self.assertTrue(Pricing(price_base=50000000).is_fix_price())
        self.assertFalse(
            Pricing(price_base=50000000, price_min=40000000).is_fix_price()
        )
        self.assertFalse(
            Pricing(price_base=50000000, price_per_volume=400).is_fix_price()
        )
        self.assertFalse(
            Pricing(price_base=50000000, price_per_collateral_percent=2).is_fix_price()
        )
        self.assertFalse(Pricing().is_fix_price())

    def test_clean_normal(self):
        p = Pricing(price_base=50000000)
        p.clean()


class TestPricingPricePerVolumeModifier(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.handler, _ = create_contract_handler_w_contracts()

    def test_return_none_if_not_set(self):
        p = Pricing()
        self.assertIsNone(p.price_per_volume_modifier())
        self.assertIsNone(p.price_per_volume_eff())

    def test_is_ignored_in_price_calculation_if_not_set(self):
        p = Pricing()
        p.price_per_volume = 50
        self.assertEqual(p.get_calculated_price(10, None), 500)

    def test_returns_none_if_not_set_in_pricing(self):
        self.handler.price_per_volume_modifier = 10
        self.handler.save()
        p = Pricing()
        p.price_per_volume = 50

        self.assertIsNone(p.price_per_volume_modifier())

    def test_can_calculate_with_plus_value(self):
        self.handler.price_per_volume_modifier = 10
        self.handler.save()

        p = Pricing()
        p.price_per_volume = 50
        p.use_price_per_volume_modifier = True

        self.assertEqual(p.price_per_volume_eff(), 55)
        self.assertEqual(p.get_calculated_price(10, None), 550)

    def test_can_calculate_with_negative_value(self):
        self.handler.price_per_volume_modifier = -10
        self.handler.save()

        p = Pricing()
        p.price_per_volume = 50
        p.use_price_per_volume_modifier = True

        self.assertEqual(p.price_per_volume_eff(), 45)
        self.assertEqual(p.get_calculated_price(10, None), 450)

    def test_calculated_price_is_never_negative(self):
        self.handler.price_per_volume_modifier = -200
        self.handler.save()

        p = Pricing()
        p.price_per_volume = 50
        p.use_price_per_volume_modifier = True

        self.assertEqual(p.price_per_volume_eff(), 0)

    def test_returns_none_if_not_set_for_handler(self):
        p = Pricing(price_base=50000000)
        p.use_price_per_volume_modifier = True
        self.assertIsNone(p.price_per_volume_modifier())

    def test_returns_none_if_no_handler_defined(self):
        ContractHandler.objects.all().delete()
        p = Pricing(price_base=50000000)
        p.use_price_per_volume_modifier = True
        self.assertIsNone(p.price_per_volume_modifier())



# Generated by Django 4.0.10 on 2023-10-30 22:12

import django.core.validators
import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("authentication", "0021_alter_userprofile_language"),
        ("eveonline", "0017_alliance_and_corp_names_are_not_unique"),
        ("freight", "0002_alter_contracthandler_last_error"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="contract",
            options={"verbose_name": "contract", "verbose_name_plural": "contracts"},
        ),
        migrations.AlterModelOptions(
            name="contractcustomernotification",
            options={
                "verbose_name": "contract customer notification",
                "verbose_name_plural": "contract customer notifications",
            },
        ),
        migrations.AlterModelOptions(
            name="contracthandler",
            options={
                "verbose_name": "contract handler",
                "verbose_name_plural": "contract handler",
            },
        ),
        migrations.AlterModelOptions(
            name="location",
            options={"verbose_name": "location", "verbose_name_plural": "locations"},
        ),
        migrations.AlterModelOptions(
            name="pricing",
            options={"verbose_name": "pricing", "verbose_name_plural": "pricings"},
        ),
        migrations.AlterField(
            model_name="contract",
            name="acceptor",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Character of acceptor or None if accepted by corp",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_acceptor",
                to="eveonline.evecharacter",
                verbose_name="acceptor",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="acceptor_corporation",
            field=models.ForeignKey(
                blank=True,
                default=None,
                help_text="Corporation of acceptor",
                null=True,
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_acceptor_corporation",
                to="eveonline.evecorporationinfo",
                verbose_name="acceptor corporation",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="collateral",
            field=models.FloatField(verbose_name="collateral"),
        ),
        migrations.AlterField(
            model_name="contract",
            name="contract_id",
            field=models.IntegerField(verbose_name="contract ID"),
        ),
        migrations.AlterField(
            model_name="contract",
            name="date_accepted",
            field=models.DateTimeField(
                blank=True, default=None, null=True, verbose_name="date accepted"
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="date_completed",
            field=models.DateTimeField(
                blank=True, default=None, null=True, verbose_name="date completed"
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="date_expired",
            field=models.DateTimeField(verbose_name="date expired"),
        ),
        migrations.AlterField(
            model_name="contract",
            name="date_issued",
            field=models.DateTimeField(verbose_name="date issues"),
        ),
        migrations.AlterField(
            model_name="contract",
            name="date_notified",
            field=models.DateTimeField(
                blank=True,
                db_index=True,
                default=None,
                help_text="Datetime of latest notification, None = none has been sent",
                null=True,
                verbose_name="date notified",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="days_to_complete",
            field=models.IntegerField(verbose_name="days to complete"),
        ),
        migrations.AlterField(
            model_name="contract",
            name="end_location",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_end_location",
                to="freight.location",
                verbose_name="end location",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="for_corporation",
            field=models.BooleanField(verbose_name="for corporation"),
        ),
        migrations.AlterField(
            model_name="contract",
            name="issuer",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_issuer",
                to="eveonline.evecharacter",
                verbose_name="issuer",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="issuer_corporation",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_issuer_corporation",
                to="eveonline.evecorporationinfo",
                verbose_name="issuer corporation",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="issues",
            field=models.TextField(
                blank=True,
                default=None,
                help_text="List or price check issues as JSON array of strings or None",
                null=True,
                verbose_name="issues",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="pricing",
            field=models.ForeignKey(
                blank=True,
                default=None,
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="contracts",
                to="freight.pricing",
                verbose_name="pricing",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="reward",
            field=models.FloatField(verbose_name="reward"),
        ),
        migrations.AlterField(
            model_name="contract",
            name="start_location",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="contracts_start_location",
                to="freight.location",
                verbose_name="start location",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="status",
            field=models.CharField(
                choices=[
                    ("outstanding", "outstanding"),
                    ("in_progress", "in progress"),
                    ("finished_issuer", "finished issuer"),
                    ("finished_contractor", "finished contractor"),
                    ("finished", "finished"),
                    ("canceled", "canceled"),
                    ("rejected", "rejected"),
                    ("failed", "failed"),
                    ("deleted", "deleted"),
                    ("reversed", "reversed"),
                ],
                db_index=True,
                max_length=32,
                verbose_name="status",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="title",
            field=models.CharField(
                blank=True,
                default=None,
                max_length=100,
                null=True,
                verbose_name="title",
            ),
        ),
        migrations.AlterField(
            model_name="contract",
            name="volume",
            field=models.FloatField(verbose_name="volume"),
        ),
        migrations.AlterField(
            model_name="contractcustomernotification",
            name="contract",
            field=models.ForeignKey(
                on_delete=django.db.models.deletion.CASCADE,
                related_name="customer_notifications",
                to="freight.contract",
                verbose_name="contract",
            ),
        ),
        migrations.AlterField(
            model_name="contractcustomernotification",
            name="date_notified",
            field=models.DateTimeField(
                help_text="datetime of notification", verbose_name="date notified"
            ),
        ),
        migrations.AlterField(
            model_name="contractcustomernotification",
            name="status",
            field=models.CharField(
                choices=[
                    ("outstanding", "outstanding"),
                    ("in_progress", "in progress"),
                    ("finished_issuer", "finished issuer"),
                    ("finished_contractor", "finished contractor"),
                    ("finished", "finished"),
                    ("canceled", "canceled"),
                    ("rejected", "rejected"),
                    ("failed", "failed"),
                    ("deleted", "deleted"),
                    ("reversed", "reversed"),
                ],
                db_index=True,
                max_length=32,
                verbose_name="status",
            ),
        ),
        migrations.AlterField(
            model_name="contracthandler",
            name="character",
            field=models.ForeignKey(
                default=None,
                help_text="Character used for syncing contracts",
                null=True,
                on_delete=django.db.models.deletion.SET_DEFAULT,
                related_name="+",
                to="authentication.characterownership",
                verbose_name="character",
            ),
        ),
        migrations.AlterField(
            model_name="contracthandler",
            name="last_sync",
            field=models.DateTimeField(
                blank=True,
                default=None,
                help_text="when the last sync happened",
                null=True,
                verbose_name="last sync at",
            ),
        ),
        migrations.AlterField(
            model_name="contracthandler",
            name="operation_mode",
            field=models.CharField(
                default="my_alliance",
                help_text="Defines what kind of contracts are synced",
                max_length=32,
                verbose_name="operation mode",
            ),
        ),
        migrations.AlterField(
            model_name="contracthandler",
            name="organization",
            field=models.OneToOneField(
                on_delete=django.db.models.deletion.CASCADE,
                primary_key=True,
                serialize=False,
                to="freight.eveentity",
                verbose_name="organization",
            ),
        ),
        migrations.AlterField(
            model_name="contracthandler",
            name="price_per_volume_modifier",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="Global modifier for price per volume in percent, e.g. 2.5 = +2.5%",
                null=True,
                verbose_name="price per volume modifier",
            ),
        ),
        migrations.AlterField(
            model_name="contracthandler",
            name="version_hash",
            field=models.CharField(
                blank=True,
                default=None,
                help_text="hash to identify changes to contracts",
                max_length=32,
                null=True,
                verbose_name="version hash",
            ),
        ),
        migrations.AlterField(
            model_name="eveentity",
            name="category",
            field=models.CharField(
                choices=[
                    ("alliance", "Alliance"),
                    ("corporation", "Corporation"),
                    ("character", "Character"),
                ],
                max_length=32,
                verbose_name="category",
            ),
        ),
        migrations.AlterField(
            model_name="eveentity",
            name="id",
            field=models.IntegerField(
                primary_key=True,
                serialize=False,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="id",
            ),
        ),
        migrations.AlterField(
            model_name="eveentity",
            name="name",
            field=models.CharField(max_length=254, verbose_name="name"),
        ),
        migrations.AlterField(
            model_name="location",
            name="id",
            field=models.BigIntegerField(
                help_text="Eve Online location ID, either item ID for stations or structure ID for structures",
                primary_key=True,
                serialize=False,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="id",
            ),
        ),
        migrations.AlterField(
            model_name="location",
            name="name",
            field=models.CharField(
                db_index=True,
                help_text="In-game name of this station or structure",
                max_length=100,
                verbose_name="name",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="collateral_max",
            field=models.BigIntegerField(
                blank=True,
                default=None,
                help_text="Maximum allowed collateral in ISK",
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="collateral maximum",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="collateral_min",
            field=models.BigIntegerField(
                blank=True,
                default=None,
                help_text="Minimum required collateral in ISK",
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="collateral minimum",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="days_to_complete",
            field=models.PositiveIntegerField(
                blank=True,
                default=None,
                help_text="Recommended days for contract completion",
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="days to complete",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="days_to_expire",
            field=models.PositiveIntegerField(
                blank=True,
                default=None,
                help_text="Recommended days for contracts to expire",
                null=True,
                validators=[django.core.validators.MinValueValidator(1)],
                verbose_name="days to expire",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="details",
            field=models.TextField(
                blank=True,
                default=None,
                help_text="Text with additional instructions for using this pricing",
                null=True,
                verbose_name="details",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="end_location",
            field=models.ForeignKey(
                help_text="Destination station or structure for courier route",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="freight.location",
                verbose_name="end location",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="is_active",
            field=models.BooleanField(
                default=True,
                help_text="Disabled pricings will not be used or shown",
                verbose_name="is active",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="is_bidirectional",
            field=models.BooleanField(
                default=True,
                help_text="Whether this pricing is valid for contracts in either direction or only the one specified",
                verbose_name="is bidirectional",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="is_default",
            field=models.BooleanField(
                default=False,
                help_text="The default pricing will be preselected in the calculator. Please make sure to only mark one pricing as default.",
                verbose_name="is default",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="price_base",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="Base price in ISK",
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="price base",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="price_min",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="Minimum total price in ISK",
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="price minimum",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="price_per_collateral_percent",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="Add-on price in percent of collateral",
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="price markup from collateral",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="price_per_volume",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="Add-on price per m3 volume in ISK",
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="price markup from volume",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="start_location",
            field=models.ForeignKey(
                help_text="Starting station or structure for courier route",
                on_delete=django.db.models.deletion.CASCADE,
                related_name="+",
                to="freight.location",
                verbose_name="start location",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="use_price_per_volume_modifier",
            field=models.BooleanField(
                default=False,
                help_text="Whether the global price per volume modifier is used",
                verbose_name="price per volume modifier enabled",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="volume_max",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="Maximum allowed volume in m3",
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="volume maximum",
            ),
        ),
        migrations.AlterField(
            model_name="pricing",
            name="volume_min",
            field=models.FloatField(
                blank=True,
                default=None,
                help_text="Minimum allowed volume in m3",
                null=True,
                validators=[django.core.validators.MinValueValidator(0)],
                verbose_name="volume minimum",
            ),
        ),
    ]



from allianceauth.eveonline.models import EveCharacter
from app_utils.testing import NoSocketsTestCase

from freight.helpers import update_or_create_eve_entity_from_evecharacter
from freight.models import EveEntity

from .testdata.helpers import load_eve_characters


class TestCreateEveEntityFromEveCharacter(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_eve_characters()
        cls.character = EveCharacter.objects.get(character_id=90000001)

    def test_can_create_corporation_from_evecharacter(self):
        corporation, _ = update_or_create_eve_entity_from_evecharacter(
            self.character, category=EveEntity.CATEGORY_CORPORATION
        )
        self.assertEqual(int(corporation.id), 92000001)

    def test_can_create_alliance_from_evecharacter(self):
        alliance, _ = update_or_create_eve_entity_from_evecharacter(
            self.character, category=EveEntity.CATEGORY_ALLIANCE
        )
        self.assertEqual(int(alliance.id), 93000001)

    def test_can_create_character_alliance_from_evecharacter(self):
        char2, _ = update_or_create_eve_entity_from_evecharacter(
            self.character, category=EveEntity.CATEGORY_CHARACTER
        )
        self.assertEqual(int(char2.id), 90000001)

    def test_raises_exception_when_trying_to_create_alliance_from_non_member(self):
        character = EveCharacter.objects.get(character_id=90000005)
        with self.assertRaises(ValueError):
            update_or_create_eve_entity_from_evecharacter(
                character, category=EveEntity.CATEGORY_ALLIANCE
            )

    def test_raises_exception_when_trying_to_create_invalid_category_from_evechar(self):
        with self.assertRaises(ValueError):
            update_or_create_eve_entity_from_evecharacter(
                self.character, category="xxx"
            )



import re

from django.test import override_settings
from django.urls import reverse
from django_webtest import WebTest

from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import NoSocketsTestCase

from freight.models import Contract, Location, Pricing

from .testdata.factories import create_pricing
from .testdata.helpers import create_contract_handler_w_contracts

_RE_COMBINE_WHITESPACE = re.compile(r"\s+")


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestCalculatorWeb(WebTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        _, cls.user = create_contract_handler_w_contracts()
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "freight.use_calculator", cls.user
        )

        jita = Location.objects.get(id=60003760)
        amamake = Location.objects.get(id=1022167642188)
        amarr = Location.objects.get(id=60008494)
        cls.pricing_1 = create_pricing(
            start_location=jita,
            end_location=amamake,
            price_base=50000000,
            price_per_volume=150,
            price_per_collateral_percent=2,
            collateral_max=5000000000,
            volume_max=320000,
            days_to_complete=3,
            days_to_expire=7,
        )
        cls.pricing_2 = create_pricing(
            start_location=jita, end_location=amarr, price_base=100000000
        )
        Contract.objects.update_pricing()

    def _calculate_price(self, pricing: Pricing, volume=None, collateral=None) -> tuple:
        """Performs a full price query with the calculator

        returns tuple of price_str, form, request
        """
        self.app.set_user(self.user)
        # load page and get our form
        response = self.app.get(reverse("freight:calculator"))
        form = response.forms["form_calculator"]

        # enter these values into form
        form["pricing"] = pricing.pk
        if volume:
            form["volume"] = volume
        if collateral:
            form["collateral"] = collateral

        # submit form and get response
        response = form.submit()
        form = response.forms["form_calculator"]

        # extract the price string
        price_str = response.html.find(id="text_price_2").string
        price_str = _RE_COMBINE_WHITESPACE.sub(
            " ", price_str
        ).strip()  # remove whitespaces to one
        return price_str, form, response

    def test_can_calculate_pricing_1(self):
        price_str, _, _ = self._calculate_price(self.pricing_1, 50000, 2000000000)
        expected = "98,000,000 ISK"
        self.assertEqual(price_str, expected)

    def test_can_calculate_pricing_2(self):
        price_str, _, _ = self._calculate_price(self.pricing_2)
        expected = "100,000,000 ISK"
        self.assertEqual(price_str, expected)

    def test_aborts_on_missing_collateral(self):
        price_str, form, _ = self._calculate_price(self.pricing_1, 50000)
        expected = "- ISK"
        self.assertEqual(price_str, expected)
        self.assertIn("Issues", form.text)
        self.assertIn("collateral is required", form.text)

    def test_aborts_on_missing_volume(self):
        price_str, form, _ = self._calculate_price(self.pricing_1, None, 500000)
        expected = "- ISK"
        self.assertEqual(price_str, expected)
        self.assertIn("Issues", form.text)
        self.assertIn("volume is required", form.text)

    def test_aborts_on_too_high_volume(self):
        price_str, form, _ = self._calculate_price(self.pricing_1, 400000, 500000)
        expected = "- ISK"
        self.assertEqual(price_str, expected)
        self.assertIn("Issues", form.text)
        self.assertIn("exceeds the maximum allowed volume", form.text)

    def test_aborts_on_too_high_collateral(self):
        price_str, form, _ = self._calculate_price(self.pricing_1, 40000, 6000000000)
        expected = "- ISK"
        self.assertEqual(price_str, expected)
        self.assertIn("Issues", form.text)
        self.assertIn("exceeds the maximum allowed collateral", form.text)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestCalculatorWeb2(WebTest):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        _, cls.user = create_contract_handler_w_contracts()
        cls.user = AuthUtils.add_permission_to_user_by_name(
            "freight.use_calculator", cls.user
        )

    def test_can_handle_no_pricing(self):
        # given
        self.app.set_user(self.user)
        # when
        response = self.app.get(reverse("freight:calculator"))
        # then
        self.assertEqual(response.status_code, 200)
        self.assertIn("Please define a pricing/route!", response.text)


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
class TestPricingSave(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        _, cls.user = create_contract_handler_w_contracts([149409016])

    def test_pricing_save_handler(self):
        # given
        jita = Location.objects.get(id=60003760)
        amamake = Location.objects.get(id=1022167642188)
        # when
        pricing = Pricing.objects.create(
            start_location=jita, end_location=amamake, price_base=500000000
        )
        # then
        contract_1 = Contract.objects.get(contract_id=149409016)
        self.assertEqual(contract_1.pricing, pricing)



from datetime import datetime, timedelta
from unittest.mock import Mock, patch

from bravado.exception import HTTPForbidden, HTTPNotFound

from django.utils.timezone import now, utc

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.eveonline.providers import ObjectNotFound
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.django import app_labels
from app_utils.testing import (
    BravadoOperationStub,
    BravadoResponseStub,
    NoSocketsTestCase,
    add_character_to_user_2,
    add_new_token,
    generate_invalid_pk,
)

from freight.models import Contract, EveEntity, Location, Pricing

from .testdata.factories import create_pricing
from .testdata.helpers import (
    characters_data,
    create_contract_handler_w_contracts,
    create_locations,
    structures_data,
)

MANAGERS_PATH = "freight.managers"
MODELS_PATH = "freight.models"


class TestEveEntityManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        esi_data = {}
        for character in characters_data:
            esi_data[character["character_id"]] = {
                "id": character["character_id"],
                "category": EveEntity.CATEGORY_CHARACTER,
                "name": character["character_name"],
            }
            esi_data[character["corporation_id"]] = {
                "id": character["corporation_id"],
                "category": EveEntity.CATEGORY_CORPORATION,
                "name": character["corporation_name"],
            }
            esi_data[character["alliance_id"]] = {
                "id": character["alliance_id"],
                "category": EveEntity.CATEGORY_ALLIANCE,
                "name": character["alliance_name"],
            }
            EveCharacter.objects.create(**character)

        cls.esi_data = esi_data
        cls.character = EveCharacter.objects.get(character_id=90000001)

    @classmethod
    def esi_post_universe_names(cls, *args, **kwargs) -> list:
        data = []
        if "ids" not in kwargs:
            raise ValueError("missing parameter: ids")
        for id in kwargs["ids"]:
            if id in cls.esi_data:
                data.append(cls.esi_data[id])

        return BravadoOperationStub(data)

    @patch(MANAGERS_PATH + ".esi")
    def test_can_create_entity(self, mock_esi):
        mock_esi.client.Universe.post_universe_names.side_effect = (
            TestEveEntityManager.esi_post_universe_names
        )

        obj, created = EveEntity.objects.update_or_create_esi(id=90000001)
        self.assertTrue(created)
        self.assertEqual(obj.id, 90000001)
        self.assertEqual(obj.name, "Bruce Wayne")
        self.assertEqual(obj.category, EveEntity.CATEGORY_CHARACTER)

    @patch(MANAGERS_PATH + ".esi")
    def test_can_create_entity_when_not_found(self, mock_esi):
        mock_esi.client.Universe.post_universe_names.side_effect = (
            TestEveEntityManager.esi_post_universe_names
        )

        obj, created = EveEntity.objects.get_or_create_esi(id=90000001)
        self.assertTrue(created)
        self.assertEqual(obj.id, 90000001)
        self.assertEqual(obj.name, "Bruce Wayne")
        self.assertEqual(obj.category, EveEntity.CATEGORY_CHARACTER)

    @patch(MANAGERS_PATH + ".esi")
    def test_can_update_entity(self, mock_esi):
        mock_esi.client.Universe.post_universe_names.side_effect = (
            TestEveEntityManager.esi_post_universe_names
        )
        obj, _ = EveEntity.objects.update_or_create_esi(id=90000001)
        obj.name = "Blue Company"
        obj.category = EveEntity.CATEGORY_CORPORATION

        obj, created = EveEntity.objects.update_or_create_esi(id=90000001)
        self.assertFalse(created)
        self.assertEqual(obj.id, 90000001)
        self.assertEqual(obj.name, "Bruce Wayne")
        self.assertEqual(obj.category, EveEntity.CATEGORY_CHARACTER)

    @patch(MANAGERS_PATH + ".esi")
    def test_raise_exception_if_entity_can_not_be_created(self, mock_esi):
        mock_esi.client.Universe.post_universe_names.side_effect = (
            TestEveEntityManager.esi_post_universe_names
        )

        with self.assertRaises(ObjectNotFound):
            EveEntity.objects.get_or_create_esi(id=666)


def get_universe_stations_station_id(*args, **kwargs) -> dict:
    if "station_id" not in kwargs:
        raise ValueError("missing parameter: station_id")

    station_id = str(kwargs["station_id"])
    if station_id not in structures_data:
        raise HTTPNotFound
    else:
        return BravadoOperationStub(structures_data[station_id])


def get_universe_structures_structure_id(*args, **kwargs) -> dict:
    if "structure_id" not in kwargs:
        raise ValueError("missing parameter: structure_id")

    structure_id = str(kwargs["structure_id"])
    if structure_id not in structures_data:
        raise HTTPNotFound
    else:
        return BravadoOperationStub(structures_data[structure_id])


@patch(MANAGERS_PATH + ".esi")
class TestLocationManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        user = AuthUtils.create_user("Bruce Wayne")
        character = add_character_to_user_2(
            user, 1001, "Bruce Wayne", 2001, "Wayne Tech"
        )
        cls.token = add_new_token(
            user, character, scopes=["esi-universe.read_structures.v1"]
        )

    def test_should_create_structure(self, mock_esi):
        # given
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            get_universe_structures_structure_id
        )
        # when
        obj, created = Location.objects.update_or_create_esi(self.token, 1000000000001)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, 1000000000001)
        self.assertEqual(obj.name, "Test Structure Alpha")
        self.assertEqual(obj.solar_system_id, 30002537)
        self.assertEqual(obj.type_id, 35832)

    def test_should_update_structure(self, mock_esi):
        # given
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            get_universe_structures_structure_id
        )
        obj, _ = Location.objects.update_or_create_esi(self.token, 1000000000001)
        obj.name = "Not my structure"
        obj.solar_system_id = 123
        obj.type_id = 456
        obj.save()
        # when
        obj, created = Location.objects.update_or_create_esi(self.token, 1000000000001)
        # then
        self.assertFalse(created)
        self.assertEqual(obj.id, 1000000000001)
        self.assertEqual(obj.name, "Test Structure Alpha")
        self.assertEqual(obj.solar_system_id, 30002537)
        self.assertEqual(obj.type_id, 35832)

    def test_should_get_existing_location(self, mock_esi):
        # given
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            get_universe_structures_structure_id
        )
        obj_created, _ = Location.objects.update_or_create_esi(
            self.token, 1000000000001
        )
        # when
        obj, created = Location.objects.get_or_create_esi(self.token, 1000000000001)
        # then
        self.assertFalse(created)
        self.assertEqual(obj, obj_created)

    def test_should_create_not_existing_location(self, mock_esi):
        # given
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            get_universe_structures_structure_id
        )
        # when
        obj, created = Location.objects.get_or_create_esi(self.token, 1000000000001)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, 1000000000001)
        self.assertEqual(obj.name, "Test Structure Alpha")
        self.assertEqual(obj.solar_system_id, 30002537)
        self.assertEqual(obj.type_id, 35832)

    def test_should_propagate_http_error_on_structure_create(self, mock_esi):
        # given
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            HTTPForbidden(BravadoResponseStub(status_code=403, reason="test"))
        )
        # when/then
        with self.assertRaises(HTTPForbidden):
            Location.objects.update_or_create_esi(self.token, 42, add_unknown=False)

    def test_should_propagates_exceptions_on_structure_create(self, mock_esi):
        # given
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            RuntimeError
        )
        # when/then
        with self.assertRaises(RuntimeError):
            Location.objects.update_or_create_esi(self.token, 42, add_unknown=False)

    def test_should_create_skeleton_structure_on_http_error_if_requested(
        self, mock_esi
    ):
        # given
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            HTTPForbidden(BravadoResponseStub(status_code=403, reason="test"))
        )
        # when
        obj, created = Location.objects.update_or_create_esi(
            self.token, 42, add_unknown=True
        )
        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, 42)

    def test_should_creates_skeleton_structure_on_exceptions_if_requested(
        self, mock_esi
    ):
        # given
        mock_esi.client.Universe.get_universe_structures_structure_id.side_effect = (
            RuntimeError
        )
        # when/then
        with self.assertRaises(RuntimeError):
            Location.objects.update_or_create_esi(self.token, 42, add_unknown=True)

    def test_should_create_station_from_scratch(self, mock_esi):
        # given
        mock_esi.client.Universe.get_universe_stations_station_id.side_effect = (
            get_universe_stations_station_id
        )
        # when
        obj, created = Location.objects.update_or_create_esi(self.token, 60000001)
        # then
        self.assertTrue(created)
        self.assertEqual(obj.id, 60000001)
        self.assertEqual(obj.name, "Test Station Charlie")
        self.assertEqual(obj.solar_system_id, 30002537)
        self.assertEqual(obj.type_id, 99)

    def test_should_update_existing_station(self, mock_esi):
        # given
        mock_esi.client.Universe.get_universe_stations_station_id.side_effect = (
            get_universe_stations_station_id
        )
        obj, created = Location.objects.update_or_create_esi(self.token, 60000001)
        obj.name = "Not my station"
        obj.solar_system_id = 123
        obj.type_id = 456
        obj.save()
        # when
        obj, created = Location.objects.update_or_create_esi(self.token, 60000001)
        # then
        self.assertFalse(created)
        self.assertEqual(obj.id, 60000001)
        self.assertEqual(obj.name, "Test Station Charlie")
        self.assertEqual(obj.solar_system_id, 30002537)
        self.assertEqual(obj.type_id, 99)

    def test_should_propagate_http_error_on_station_create(self, mock_esi):
        # given
        mock_esi.client.Universe.get_universe_stations_station_id.side_effect = (
            HTTPNotFound(BravadoResponseStub(status_code=404, reason="test"))
        )
        # when/then
        with self.assertRaises(HTTPNotFound):
            Location.objects.update_or_create_esi(
                self.token, 60000001, add_unknown=False
            )


class TestContractQuerySet(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.handler, cls.user = create_contract_handler_w_contracts(
            [149409016, 149409061, 149409062, 149409063, 149409064, 149409006]
        )

    def test_pending_count(self):
        result = Contract.objects.all().pending_count()
        self.assertEqual(result, 6)


class TestContractManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.handler, cls.user = create_contract_handler_w_contracts(
            [
                149409016,
                149409061,
                149409062,
                149409063,
                149409064,
                149409006,
                149409318,
            ]
        )

    def test_issued_by_user(self):
        # when
        qs = Contract.objects.all().issued_by_user(user=self.user)
        # then
        self.assertSetEqual(
            set(qs.values_list("contract_id", flat=True)),
            {149409016, 149409061, 149409062, 149409063, 149409064},
        )

    def test_can_update_pricing_for_bidirectional(self):
        # given
        jita = Location.objects.get(id=60003760)
        amamake = Location.objects.get(id=1022167642188)
        amarr = Location.objects.get(id=60008494)
        pricing_1 = create_pricing(
            start_location=jita,
            end_location=amamake,
            price_base=500000000,
            is_bidirectional=True,
        )
        pricing_3 = create_pricing(
            start_location=amarr,
            end_location=amamake,
            price_base=250000000,
            is_bidirectional=True,
        )
        # when
        result = Contract.objects.update_pricing()
        # then
        self.assertEqual(result, 7)
        contract_1 = Contract.objects.get(contract_id=149409016)
        self.assertEqual(contract_1.pricing, pricing_1)
        contract_2 = Contract.objects.get(contract_id=149409061)
        self.assertEqual(contract_2.pricing, pricing_1)
        contract_3 = Contract.objects.get(contract_id=149409062)
        self.assertEqual(contract_3.pricing, pricing_3)

    def test_can_update_pricing_for_unidirectional(self):
        # given
        jita = Location.objects.get(id=60003760)
        amamake = Location.objects.get(id=1022167642188)
        amarr = Location.objects.get(id=60008494)
        pricing_1 = create_pricing(
            start_location=jita,
            end_location=amamake,
            price_base=500000000,
            is_bidirectional=False,
        )
        pricing_2 = create_pricing(
            start_location=amamake,
            end_location=jita,
            price_base=350000000,
            is_bidirectional=False,
        )
        pricing_3 = create_pricing(
            start_location=amarr,
            end_location=amamake,
            price_base=250000000,
            is_bidirectional=True,
        )
        # when
        Contract.objects.update_pricing()
        # then
        contract_1 = Contract.objects.get(contract_id=149409016)
        self.assertEqual(contract_1.pricing, pricing_1)
        contract_2 = Contract.objects.get(contract_id=149409061)
        self.assertEqual(contract_2.pricing, pricing_2)
        contract_3 = Contract.objects.get(contract_id=149409062)
        self.assertEqual(contract_3.pricing, pricing_3)
        contract_4 = Contract.objects.get(contract_id=149409063)
        self.assertEqual(contract_4.pricing, pricing_3)
        contract_5 = Contract.objects.get(contract_id=149409064)
        self.assertIsNone(contract_5.pricing)


class TestContractManagerCreateFromDict(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.handler, cls.user = create_contract_handler_w_contracts(
            [149409016, 149409061, 149409062, 149409063, 149409064]
        )

    def test_can_create_outstanding(self):
        contract_dict = {
            "acceptor_id": 0,
            "assignee_id": 93000001,
            "availability": "personal",
            "buyout": None,
            "collateral": 50000000.0,
            "contract_id": 149409014,
            "date_accepted": None,
            "date_completed": None,
            "date_expired": datetime(2019, 10, 30, 23, tzinfo=utc),
            "date_issued": datetime(2019, 10, 2, 23, tzinfo=utc),
            "days_to_complete": 3,
            "end_location_id": 1022167642188,
            "for_corporation": False,
            "issuer_corporation_id": 92000002,
            "issuer_id": 90000003,
            "price": 0.0,
            "reward": 25000000.0,
            "start_location_id": 60003760,
            "status": "outstanding",
            "title": "demo contract",
            "type": "courier",
            "volume": 115000.0,
        }
        obj, created = Contract.objects.update_or_create_from_dict(
            self.handler, contract_dict, Mock()
        )
        self.assertTrue(created)
        self.assertEqual(obj.contract_id, 149409014)
        self.assertIsNone(obj.acceptor)
        self.assertIsNone(obj.acceptor_corporation)
        self.assertEqual(obj.collateral, 50000000)
        self.assertIsNone(obj.date_accepted)
        self.assertIsNone(obj.date_completed)
        self.assertEqual(obj.date_expired, datetime(2019, 10, 30, 23, tzinfo=utc))
        self.assertEqual(obj.date_issued, datetime(2019, 10, 2, 23, tzinfo=utc))
        self.assertEqual(obj.days_to_complete, 3)
        self.assertEqual(obj.end_location_id, 1022167642188)
        self.assertFalse(obj.for_corporation)
        self.assertEqual(
            obj.issuer_corporation,
            EveCorporationInfo.objects.get(corporation_id=92000002),
        )
        self.assertEqual(obj.issuer, EveCharacter.objects.get(character_id=90000003))
        self.assertEqual(obj.reward, 25000000)
        self.assertEqual(obj.start_location_id, 60003760)
        self.assertEqual(obj.status, Contract.Status.OUTSTANDING)
        self.assertEqual(obj.title, "demo contract")
        self.assertEqual(obj.volume, 115000)
        self.assertIsNone(obj.pricing)
        self.assertIsNone(obj.issues)

    def test_can_create_in_progress(self):
        contract_dict = {
            "acceptor_id": 90000003,
            "assignee_id": 90000003,
            "availability": "personal",
            "buyout": None,
            "collateral": 50000000.0,
            "contract_id": 149409014,
            "date_accepted": datetime(2019, 10, 3, 23, tzinfo=utc),
            "date_completed": None,
            "date_expired": datetime(2019, 10, 30, 23, tzinfo=utc),
            "date_issued": datetime(2019, 10, 2, 23, tzinfo=utc),
            "days_to_complete": 3,
            "end_location_id": 1022167642188,
            "for_corporation": False,
            "issuer_corporation_id": 92000002,
            "issuer_id": 90000003,
            "price": 0.0,
            "reward": 25000000.0,
            "start_location_id": 60003760,
            "status": "in_progress",
            "title": "demo contract",
            "type": "courier",
            "volume": 115000.0,
        }
        obj, created = Contract.objects.update_or_create_from_dict(
            self.handler, contract_dict, Mock()
        )
        self.assertTrue(created)
        self.assertEqual(obj.contract_id, 149409014)
        self.assertEqual(obj.acceptor, EveCharacter.objects.get(character_id=90000003))
        self.assertEqual(
            obj.acceptor_corporation,
            EveCorporationInfo.objects.get(corporation_id=92000002),
        )
        self.assertEqual(obj.collateral, 50000000)
        self.assertEqual(obj.date_accepted, datetime(2019, 10, 3, 23, tzinfo=utc))
        self.assertIsNone(obj.date_completed)
        self.assertEqual(obj.date_issued, datetime(2019, 10, 2, 23, tzinfo=utc))
        self.assertEqual(obj.date_expired, datetime(2019, 10, 30, 23, tzinfo=utc))
        self.assertEqual(obj.days_to_complete, 3)
        self.assertEqual(obj.end_location_id, 1022167642188)
        self.assertFalse(obj.for_corporation)
        self.assertEqual(
            obj.issuer_corporation,
            EveCorporationInfo.objects.get(corporation_id=92000002),
        )
        self.assertEqual(obj.issuer, EveCharacter.objects.get(character_id=90000003))
        self.assertEqual(obj.reward, 25000000)
        self.assertEqual(obj.start_location_id, 60003760)
        self.assertEqual(obj.status, Contract.Status.IN_PROGRESS)
        self.assertEqual(obj.title, "demo contract")
        self.assertEqual(obj.volume, 115000)
        self.assertIsNone(obj.pricing)
        self.assertIsNone(obj.issues)

    def test_can_create_finished(self):
        contract_dict = {
            "acceptor_id": 90000003,
            "assignee_id": 90000003,
            "availability": "personal",
            "buyout": None,
            "collateral": 50000000.0,
            "contract_id": 149409014,
            "date_accepted": datetime(2019, 10, 3, 23, tzinfo=utc),
            "date_completed": datetime(2019, 10, 4, 23, tzinfo=utc),
            "date_expired": datetime(2019, 10, 30, 23, tzinfo=utc),
            "date_issued": datetime(2019, 10, 2, 23, tzinfo=utc),
            "days_to_complete": 3,
            "end_location_id": 1022167642188,
            "for_corporation": False,
            "issuer_corporation_id": 92000002,
            "issuer_id": 90000003,
            "price": 0.0,
            "reward": 25000000.0,
            "start_location_id": 60003760,
            "status": "finished",
            "title": "demo contract",
            "type": "courier",
            "volume": 115000.0,
        }
        obj, created = Contract.objects.update_or_create_from_dict(
            self.handler, contract_dict, Mock()
        )
        self.assertTrue(created)
        self.assertEqual(obj.contract_id, 149409014)
        self.assertEqual(obj.acceptor, EveCharacter.objects.get(character_id=90000003))
        self.assertEqual(
            obj.acceptor_corporation,
            EveCorporationInfo.objects.get(corporation_id=92000002),
        )
        self.assertEqual(obj.collateral, 50000000)
        self.assertEqual(obj.date_accepted, datetime(2019, 10, 3, 23, tzinfo=utc))
        self.assertEqual(obj.date_completed, datetime(2019, 10, 4, 23, tzinfo=utc))
        self.assertEqual(obj.date_issued, datetime(2019, 10, 2, 23, tzinfo=utc))
        self.assertEqual(obj.date_expired, datetime(2019, 10, 30, 23, tzinfo=utc))
        self.assertEqual(obj.days_to_complete, 3)
        self.assertEqual(obj.end_location_id, 1022167642188)
        self.assertFalse(obj.for_corporation)
        self.assertEqual(
            obj.issuer_corporation,
            EveCorporationInfo.objects.get(corporation_id=92000002),
        )
        self.assertEqual(obj.issuer, EveCharacter.objects.get(character_id=90000003))
        self.assertEqual(obj.reward, 25000000)
        self.assertEqual(obj.start_location_id, 60003760)
        self.assertEqual(obj.status, Contract.Status.FINISHED)
        self.assertEqual(obj.title, "demo contract")
        self.assertEqual(obj.volume, 115000)
        self.assertIsNone(obj.pricing)
        self.assertIsNone(obj.issues)

    def test_raises_exception_on_wrong_date_types(self):
        contract_dict = {
            "acceptor_id": 90000003,
            "assignee_id": 90000003,
            "availability": "personal",
            "buyout": None,
            "collateral": 50000000.0,
            "contract_id": 149409014,
            "date_accepted": "2019-10-03T23:00:00Z",
            "date_completed": "2019-10-04T23:00:00Z",
            "date_expired": "2019-10-30T23:00:00Z",
            "date_issued": "2019-10-02T23:00:00Z",
            "days_to_complete": 3,
            "end_location_id": 1022167642188,
            "for_corporation": False,
            "issuer_corporation_id": 92000002,
            "issuer_id": 90000003,
            "price": 0.0,
            "reward": 25000000.0,
            "start_location_id": 60003760,
            "status": "finished",
            "title": "demo contract",
            "type": "courier",
            "volume": 115000.0,
        }
        with self.assertRaises(TypeError):
            Contract.objects.update_or_create_from_dict(
                self.handler, contract_dict, Mock()
            )

    @patch(MANAGERS_PATH + ".EveCharacter.objects.create_character")
    def test_can_create_in_progress_and_creates_acceptor_char(
        self, mock_create_character
    ):
        def create_character(character_id):
            return EveCharacter.objects.create(
                character_id=90000987,
                character_name="Dummy",
                corporation_id=92000002,
                corporation_name="The Planet",
            )

        mock_create_character.side_effect = create_character
        EveEntity.objects.create(
            id=90000987, name="Dummy", category=EveEntity.CATEGORY_CHARACTER
        )
        contract_dict = {
            "acceptor_id": 90000987,
            "assignee_id": 90000987,
            "availability": "personal",
            "buyout": None,
            "collateral": 50000000.0,
            "contract_id": 149409014,
            "date_accepted": datetime(2019, 10, 3, 23, tzinfo=utc),
            "date_completed": None,
            "date_expired": datetime(2019, 10, 30, 23, tzinfo=utc),
            "date_issued": datetime(2019, 10, 2, 23, tzinfo=utc),
            "days_to_complete": 3,
            "end_location_id": 1022167642188,
            "for_corporation": False,
            "issuer_corporation_id": 92000002,
            "issuer_id": 90000003,
            "price": 0.0,
            "reward": 25000000.0,
            "start_location_id": 60003760,
            "status": "in_progress",
            "title": "demo contract",
            "type": "courier",
            "volume": 115000.0,
        }
        obj, created = Contract.objects.update_or_create_from_dict(
            self.handler, contract_dict, Mock()
        )
        self.assertTrue(created)
        self.assertEqual(obj.contract_id, 149409014)
        self.assertEqual(obj.acceptor, EveCharacter.objects.get(character_id=90000987))
        self.assertEqual(
            obj.acceptor_corporation,
            EveCorporationInfo.objects.get(corporation_id=92000002),
        )

    @patch(MANAGERS_PATH + ".EveEntityManager.get_or_create_esi")
    def test_sets_acceptor_to_none_if_it_cant_be_created(self, mock_get_or_create_esi):
        # given
        mock_get_or_create_esi.side_effect = OSError
        contract_dict = {
            "acceptor_id": 666,
            "assignee_id": 90000987,
            "availability": "personal",
            "buyout": None,
            "collateral": 50000000.0,
            "contract_id": 149409014,
            "date_accepted": datetime(2019, 10, 3, 23, tzinfo=utc),
            "date_completed": None,
            "date_expired": datetime(2019, 10, 30, 23, tzinfo=utc),
            "date_issued": datetime(2019, 10, 2, 23, tzinfo=utc),
            "days_to_complete": 3,
            "end_location_id": 1022167642188,
            "for_corporation": False,
            "issuer_corporation_id": 92000002,
            "issuer_id": 90000003,
            "price": 0.0,
            "reward": 25000000.0,
            "start_location_id": 60003760,
            "status": "in_progress",
            "title": "demo contract",
            "type": "courier",
            "volume": 115000.0,
        }
        # when
        obj, created = Contract.objects.update_or_create_from_dict(
            self.handler, contract_dict, Mock()
        )
        # then
        self.assertTrue(created)
        self.assertEqual(obj.contract_id, 149409014)
        self.assertIsNone(obj.acceptor)
        self.assertIsNone(obj.acceptor_corporation)


if "discord" in app_labels():

    @patch(MODELS_PATH + ".contracts.FREIGHT_HOURS_UNTIL_STALE_STATUS", 48)
    @patch(MODELS_PATH + ".contracts.dhooks_lite.Webhook.execute", autospec=True)
    class TestContractManagerNotifications(NoSocketsTestCase):
        @classmethod
        def setUpClass(cls):
            super().setUpClass()
            cls.handler, _ = create_contract_handler_w_contracts()
            jita = Location.objects.get(id=60003760)
            amamake = Location.objects.get(id=1022167642188)
            create_pricing(
                start_location=jita, end_location=amamake, price_base=500000000
            )
            Contract.objects.update_pricing()

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", False)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_send_pilot_notifications_normal(self, mock_webhook_execute):
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 8)

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", True)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_send_pilot_notifications_if_invalid_route_set_but_global_option_enabled(
            self, mock_webhook_execute
        ):
            x = Contract.objects.filter(status=Contract.Status.OUTSTANDING).first()
            Contract.objects.all().exclude(pk=x.pk).delete()
            x.end_location_id = 60008494
            x.save()
            Contract.objects.update_pricing()
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 1)

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", False)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_send_pilot_notifications_if_invalid_route_set_and_global_option_disabled(
            self, mock_webhook_execute
        ):
            x = Contract.objects.filter(status=Contract.Status.OUTSTANDING).first()
            Contract.objects.all().exclude(pk=x.pk).delete()
            x.end_location_id = 60008494
            x.save()
            Contract.objects.update_pricing()
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 0)

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", False)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_dont_send_pilot_notifications_for_expired_contracts(
            self, mock_webhook_execute
        ):
            x = Contract.objects.filter(status=Contract.Status.OUTSTANDING).first()
            Contract.objects.all().exclude(pk=x.pk).delete()
            x.date_expired = now() - timedelta(hours=1)
            x.save()
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 0)

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", False)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_send_pilot_notifications_only_once(self, mock_webhook_execute):
            x = Contract.objects.filter(status=Contract.Status.OUTSTANDING).first()
            Contract.objects.all().exclude(pk=x.pk).delete()

            # round #1
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 1)

            # round #2
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 1)

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", False)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_dont_send_any_notifications_when_no_url_if_set(
            self, mock_webhook_execute
        ):
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 0)

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", False)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_send_customer_notifications_normal(self, mock_webhook_execute):
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 12)

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", False)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_dont_send_customer_notifications_for_expired_contracts(
            self, mock_webhook_execute
        ):
            x = Contract.objects.filter(status=Contract.Status.OUTSTANDING).first()
            Contract.objects.all().exclude(pk=x.pk).delete()
            x.date_expired = now() - timedelta(hours=1)
            x.save()
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 0)

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", False)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", "url")
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_send_customer_notifications_only_once_per_state(
            self, mock_webhook_execute
        ):
            x = Contract.objects.filter(status=Contract.Status.OUTSTANDING).first()
            Contract.objects.all().exclude(pk=x.pk).delete()

            # round #1
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 1)

            # round #2
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 1)

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", True)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_send_customer_notification_if_invalid_route_set_but_global_option_enabled(
            self, mock_webhook_execute
        ):
            x = Contract.objects.filter(status=Contract.Status.OUTSTANDING).first()
            Contract.objects.all().exclude(pk=x.pk).delete()
            x.end_location_id = 60008494
            x.save()
            Contract.objects.update_pricing()
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 1)

        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MANAGERS_PATH + ".FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MANAGERS_PATH + ".FREIGHT_NOTIFY_ALL_CONTRACTS", False)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_WEBHOOK_URL", "url")
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORD_CUSTOMERS_WEBHOOK_URL", None)
        @patch(MODELS_PATH + ".contracts.FREIGHT_DISCORDPROXY_ENABLED", False)
        def test_send_customer_notification_if_invalid_route_set_and_global_option_disabled(
            self, mock_webhook_execute
        ):
            x = Contract.objects.filter(status=Contract.Status.OUTSTANDING).first()
            Contract.objects.all().exclude(pk=x.pk).delete()
            x.end_location_id = 60008494
            x.save()
            Contract.objects.update_pricing()
            Contract.objects.send_notifications(rate_limited=False)
            self.assertEqual(mock_webhook_execute.call_count, 0)


class TestPricingManager(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.jita, cls.amamake, cls.amarr = create_locations()
        cls.p1 = create_pricing(
            start_location=cls.jita,
            end_location=cls.amamake,
            price_base=50000000,
            is_default=True,
        )
        cls.p2 = create_pricing(
            start_location=cls.jita, end_location=cls.amarr, price_base=10000000
        )

    def test_default_pricing_no_default_defined(self):
        Pricing.objects.all().delete()
        p = create_pricing(
            start_location=self.jita,
            end_location=self.amamake,
            price_base=50000000,
            is_default=True,
        )
        expected = p
        self.assertEqual(Pricing.objects.get_default(), expected)

    def test_default_and_default_defined(self):
        expected = self.p1
        self.assertEqual(Pricing.objects.get_default(), expected)

    def test_default_with_no_pricing_defined(self):
        Pricing.objects.all().delete()
        expected = None
        self.assertEqual(Pricing.objects.get_default(), expected)

    def test_get_or_default_normal(self):
        expected = self.p1
        self.assertEqual(Pricing.objects.get_or_default(self.p1.pk), expected)

    def test_get_or_default_not_found(self):
        expected = self.p1
        invalid_pk = generate_invalid_pk(Pricing)
        self.assertEqual(Pricing.objects.get_or_default(invalid_pk), expected)

    def test_get_or_default_with_none(self):
        expected = self.p1
        self.assertEqual(Pricing.objects.get_or_default(None), expected)



"""Helpers for Freight."""

from typing import Any, Tuple

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo


def update_or_create_eve_entity_from_evecharacter(
    character: EveCharacter, category: str
) -> Tuple[Any, bool]:
    """Update or create an EveEntity object from an EveCharacter object."""
    from .models import EveEntity

    if category == EveEntity.CATEGORY_ALLIANCE:
        if not character.alliance_id:
            raise ValueError("character is not an alliance member")
        return EveEntity.objects.update_or_create(
            id=character.alliance_id,
            defaults={
                "name": character.alliance_name,
                "category": EveEntity.CATEGORY_ALLIANCE,
            },
        )

    if category == EveEntity.CATEGORY_CORPORATION:
        return EveEntity.objects.update_or_create(
            id=character.corporation_id,
            defaults={
                "name": character.corporation_name,
                "category": EveEntity.CATEGORY_CORPORATION,
            },
        )

    if category == EveEntity.CATEGORY_CHARACTER:
        return EveEntity.objects.update_or_create(
            id=character.character_id,
            defaults={
                "name": character.character_name,
                "category": EveEntity.CATEGORY_CHARACTER,
            },
        )

    raise ValueError(f"Invalid category: f{category}")


def get_or_create_eve_character(character_id: int) -> Tuple[Any, bool]:
    """Get or create EveCharacter object."""
    try:
        return EveCharacter.objects.get(character_id=character_id), False
    except EveCharacter.DoesNotExist:
        return EveCharacter.objects.create_character(character_id=character_id), True


def get_or_create_eve_corporation_info(corporation_id: int) -> Tuple[Any, bool]:
    """Get or create EveCorporationInfo object."""
    try:
        return (
            EveCorporationInfo.objects.get(corporation_id=corporation_id),
            False,
        )
    except EveCorporationInfo.DoesNotExist:
        return (
            EveCorporationInfo.objects.create_corporation(corp_id=corporation_id),
            True,
        )



import datetime as dt
import inspect
import json
import os
from random import randrange
from unittest.mock import Mock

from django.contrib.auth.models import User
from django.utils.timezone import now

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.django import app_labels

from freight.models import Contract, ContractHandler, EveEntity, Location

if "discord" in app_labels():
    from allianceauth.services.modules.discord.models import DiscordUser


_currentdir = os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))


def _load_structures_data() -> list:
    with open(_currentdir + "/universe_structures.json", "r", encoding="utf-8") as f:
        data = json.load(f)

    return data


def _load_characters_data() -> list:
    with open(_currentdir + "/characters.json", "r", encoding="utf-8") as f:
        data = json.load(f)

    return data


def _load_contract_data() -> list:
    with open(_currentdir + "/contracts.json", "r", encoding="utf-8") as f:
        _contracts_data = json.load(f)

    # update dates to something current, so won't be treated as stale
    for contract in _contracts_data:
        date_issued = now() - dt.timedelta(hours=12, minutes=randrange(30))
        date_accepted = date_issued + dt.timedelta(hours=2, minutes=randrange(30))
        date_completed = date_accepted + dt.timedelta(hours=3, minutes=randrange(30))
        date_expired = now() + dt.timedelta(days=7 + randrange(7), hours=randrange(10))
        if "date_issued" in contract:
            contract["date_issued"] = date_issued

        if "date_accepted" in contract:
            contract["date_accepted"] = date_accepted

        if "date_completed" in contract:
            contract["date_completed"] = date_completed

        if "date_expired" in contract:
            contract["date_expired"] = date_expired

    return _contracts_data


contracts_data = _load_contract_data()
characters_data = _load_characters_data()
structures_data = _load_structures_data()


def create_locations():
    jita = Location.objects.create(
        id=60003760,
        name="Jita IV - Moon 4 - Caldari Navy Assembly Plant",
        solar_system_id=30000142,
        type_id=52678,
        category_id=3,
    )
    amamake = Location.objects.create(
        id=1022167642188,
        name="Amamake - 3 Time Nearly AT Winners",
        solar_system_id=30002537,
        type_id=35834,
        category_id=65,
    )
    amarr = Location.objects.create(
        id=60008494,
        name="Amarr VIII (Oris) - Emperor Family Academy",
        solar_system_id=30002187,
        type_id=1932,
        category_id=3,
    )
    return jita, amamake, amarr


def load_eve_characters():
    for character in characters_data:
        EveCharacter.objects.create(**character)


def create_user_from_character(character: EveCharacter) -> User:
    user = AuthUtils.create_user(username=character.character_name)
    user.profile.main_character = character
    user.profile.save()
    return user


def create_entities_from_characters():
    for character in characters_data:
        EveCharacter.objects.create(**character)
        EveCorporationInfo.objects.get_or_create(
            corporation_id=character["corporation_id"],
            defaults={
                "corporation_name": character["corporation_name"],
                "corporation_ticker": character["corporation_ticker"],
                "member_count": 42,
            },
        )
        EveEntity.objects.get_or_create(
            id=character["character_id"],
            defaults={
                "category": EveEntity.CATEGORY_CHARACTER,
                "name": character["character_name"],
            },
        )
        EveEntity.objects.get_or_create(
            id=character["corporation_id"],
            defaults={
                "category": EveEntity.CATEGORY_CORPORATION,
                "name": character["corporation_name"],
            },
        )
        if "alliance_id" in character and character["alliance_id"] is not None:
            EveEntity.objects.get_or_create(
                id=character["alliance_id"],
                defaults={
                    "category": EveEntity.CATEGORY_ALLIANCE,
                    "name": character["alliance_name"],
                },
            )


def create_contract_handler_w_contracts(selected_contract_ids: list = None) -> tuple:
    """create contract handler with contracts and all related entities"""

    create_entities_from_characters()

    # 1 user
    my_character = EveCharacter.objects.get(character_id=90000001)
    my_organization = EveEntity.objects.get(id=my_character.alliance_id)
    User.objects.filter(username=my_character.character_name).delete()
    my_user = AuthUtils.create_user(my_character.character_name)
    my_user = AuthUtils.add_permission_to_user_by_name("freight.basic_access", my_user)
    my_main_ownership = CharacterOwnership.objects.create(
        character=my_character, owner_hash="x1", user=my_user
    )
    my_user.profile.main_character = my_character
    my_user.profile.save()
    my_user = User.objects.get(username=my_character.character_name)
    my_handler = ContractHandler.objects.create(
        organization=my_organization, character=my_main_ownership
    )

    create_locations()

    for contract in contracts_data:
        if (
            not selected_contract_ids
            or contract["contract_id"] in selected_contract_ids
        ):
            if contract["type"] == "courier":
                Contract.objects.update_or_create_from_dict(
                    handler=my_handler, contract=contract, token=Mock()
                )

    # create users and Discord accounts from contract issuers
    has_discord = "discord" in app_labels()
    for contract in Contract.objects.all():
        issuer_user = User.objects.filter(
            character_ownerships__character=contract.issuer
        ).first()
        if not issuer_user:
            issuer_user = User.objects.create_user(
                contract.issuer.character_name, "abc@example.com", "password"
            )
            CharacterOwnership.objects.create(
                character=contract.issuer,
                owner_hash=contract.issuer.character_name + "x",
                user=issuer_user,
            )

        if has_discord:
            DiscordUser.objects.update_or_create(
                user=issuer_user, defaults={"uid": contract.issuer.character_id}
            )

    return my_handler, my_user



import datetime as dt
from unittest.mock import Mock, patch

from django.contrib.auth.models import User
from django.utils.timezone import now
from esi.errors import TokenError, TokenExpiredError, TokenInvalidError
from esi.models import Token

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo
from allianceauth.tests.auth_utils import AuthUtils
from app_utils.testing import BravadoOperationStub, NoSocketsTestCase

from freight.app_settings import (
    FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE,
    FREIGHT_OPERATION_MODE_CORP_PUBLIC,
    FREIGHT_OPERATION_MODE_MY_ALLIANCE,
    FREIGHT_OPERATION_MODE_MY_CORPORATION,
    FREIGHT_OPERATION_MODES,
)
from freight.models import Contract, ContractHandler, EveEntity, Freight

from ..testdata.helpers import (
    characters_data,
    contracts_data,
    create_entities_from_characters,
    create_locations,
)

MODULE_PATH = "freight.models.contract_handlers"
PATCH_FREIGHT_OPERATION_MODE = MODULE_PATH + ".FREIGHT_OPERATION_MODE"


class TestContractHandler(NoSocketsTestCase):
    def setUp(self):
        for character in characters_data:
            EveCharacter.objects.create(**character)
            EveCorporationInfo.objects.get_or_create(
                corporation_id=character["corporation_id"],
                defaults={
                    "corporation_name": character["corporation_name"],
                    "corporation_ticker": character["corporation_ticker"],
                    "member_count": 42,
                },
            )

        # 1 user
        self.character = EveCharacter.objects.get(character_id=90000001)
        self.corporation = EveCorporationInfo.objects.get(
            corporation_id=self.character.corporation_id
        )
        self.organization = EveEntity.objects.create(
            id=self.character.alliance_id,
            category=EveEntity.CATEGORY_ALLIANCE,
            name=self.character.alliance_name,
        )
        self.user = User.objects.create_user(
            self.character.character_name, "abc@example.com", "password"
        )
        self.main_ownership = CharacterOwnership.objects.create(
            character=self.character, owner_hash="x1", user=self.user
        )
        self.handler = ContractHandler.objects.create(
            organization=self.organization, character=self.main_ownership
        )

    def test_str(self):
        self.assertEqual(str(self.handler), "Justice League")

    def test_repr(self):
        expected = (
            f"ContractHandler(pk={self.handler.pk}, organization='Justice League')"
        )
        self.assertEqual(repr(self.handler), expected)

    def test_operation_mode_friendly(self):
        self.handler.operation_mode = FREIGHT_OPERATION_MODE_MY_ALLIANCE
        self.assertEqual(self.handler.operation_mode_friendly, "My Alliance")
        self.handler.operation_mode = "undefined operation mode"
        with self.assertRaises(ValueError):
            self.handler.operation_mode_friendly

    def test_get_availability_text_for_contracts(self):
        self.handler.operation_mode = FREIGHT_OPERATION_MODE_MY_ALLIANCE
        self.assertEqual(
            self.handler.get_availability_text_for_contracts(),
            "Private (Justice League) [My Alliance]",
        )
        self.handler.operation_mode = FREIGHT_OPERATION_MODE_MY_CORPORATION
        self.assertEqual(
            self.handler.get_availability_text_for_contracts(),
            "Private (Justice League) [My Corporation]",
        )
        self.handler.operation_mode = FREIGHT_OPERATION_MODE_CORP_PUBLIC
        self.assertEqual(
            self.handler.get_availability_text_for_contracts(),
            "Private (Justice League) ",
        )

    @patch(MODULE_PATH + ".FREIGHT_CONTRACT_SYNC_GRACE_MINUTES", 30)
    def test_is_sync_ok(self):
        # recent sync
        self.handler.last_sync = now()
        self.assertTrue(self.handler.is_sync_ok)

        # sync within grace period
        self.handler.last_sync = now() - dt.timedelta(minutes=29)
        self.assertTrue(self.handler.is_sync_ok)

        # no sync within grace period
        self.handler.last_sync = now() - dt.timedelta(minutes=31)
        self.assertFalse(self.handler.is_sync_ok)


class TestContractsSync(NoSocketsTestCase):
    def setUp(self):
        create_entities_from_characters()

        # 1 user
        self.character = EveCharacter.objects.get(character_id=90000001)

        self.alliance = EveEntity.objects.get(id=self.character.alliance_id)
        self.corporation = EveEntity.objects.get(id=self.character.corporation_id)
        self.user = User.objects.create_user(
            self.character.character_name, "abc@example.com", "password"
        )
        self.main_ownership = CharacterOwnership.objects.create(
            character=self.character, owner_hash="x1", user=self.user
        )
        create_locations()

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_MY_CORPORATION)
    def test_abort_on_wrong_operation_mode(self):
        # given
        handler = ContractHandler.objects.create(
            organization=self.alliance,
            operation_mode=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
            character=self.main_ownership,
        )
        # when/then
        with self.assertRaises(ValueError):
            handler.update_contracts_esi()

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    def test_abort_when_no_sync_char(self):
        # given
        handler = ContractHandler.objects.create(
            organization=self.alliance,
            operation_mode=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
        )
        # when/Then
        with self.assertRaises(ValueError):
            handler.update_contracts_esi()

    # test expired token
    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    @patch(MODULE_PATH + ".Token")
    def test_abort_when_token_expired(self, mock_Token):
        # given
        mock_Token.objects.filter.side_effect = TokenExpiredError()
        self.user = AuthUtils.add_permission_to_user_by_name(
            "freight.setup_contract_handler", self.user
        )
        handler = ContractHandler.objects.create(
            organization=self.alliance,
            character=self.main_ownership,
            operation_mode=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
        )
        # when/then
        with self.assertRaises(TokenExpiredError):
            handler.update_contracts_esi()

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    @patch(MODULE_PATH + ".Token")
    def test_abort_when_token_invalid(self, mock_Token):
        mock_Token.objects.filter.side_effect = TokenInvalidError()
        self.user = AuthUtils.add_permission_to_user_by_name(
            "freight.setup_contract_handler", self.user
        )
        handler = ContractHandler.objects.create(
            organization=self.alliance,
            character=self.main_ownership,
            operation_mode=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
        )

        # when/then
        with self.assertRaises(TokenInvalidError):
            handler.update_contracts_esi()

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    @patch(MODULE_PATH + ".Token")
    def test_abort_when_no_token_exists(self, mock_Token):
        mock_Token.objects.filter.return_value.require_scopes.return_value.require_valid.return_value.first.return_value = (
            None
        )
        self.user = AuthUtils.add_permission_to_user_by_name(
            "freight.setup_contract_handler", self.user
        )
        handler = ContractHandler.objects.create(
            organization=self.alliance,
            character=self.main_ownership,
            operation_mode=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
        )
        with self.assertRaises(TokenError):
            handler.update_contracts_esi()

    @staticmethod
    def esi_get_corporations_corporation_id_contracts(**kwargs):
        return BravadoOperationStub(contracts_data)

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    @patch(MODULE_PATH + ".ContractHandler.contracts")
    @patch(MODULE_PATH + ".Token")
    @patch(MODULE_PATH + ".esi")
    def test_continue_when_exception_occurs_during_contract_creation(
        self,
        mock_esi,
        mock_Token,
        mock_Contracts_objects_update_or_create_from_dict,
    ):
        mock_Contracts_objects_update_or_create_from_dict.update_or_create_from_dict.side_effect = (
            OSError
        )
        mock_Contracts = mock_esi.client.Contracts
        mock_Contracts.get_corporations_corporation_id_contracts.side_effect = (
            self.esi_get_corporations_corporation_id_contracts
        )
        mock_Token.objects.filter.return_value.require_scopes.return_value.require_valid.return_value.first.return_value = Mock(
            spec=Token
        )
        self.user = AuthUtils.add_permission_to_user_by_name(
            "freight.setup_contract_handler", self.user
        )
        handler = ContractHandler.objects.create(
            organization=self.alliance,
            character=self.main_ownership,
            operation_mode=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
        )

        # when
        handler.update_contracts_esi()

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    @patch(MODULE_PATH + ".Token")
    @patch(MODULE_PATH + ".esi")
    def test_can_sync_contracts_for_my_alliance(self, mock_esi, mock_Token):
        mock_Contracts = mock_esi.client.Contracts
        mock_Contracts.get_corporations_corporation_id_contracts.side_effect = (
            self.esi_get_corporations_corporation_id_contracts
        )
        mock_Token.objects.filter.return_value.require_scopes.return_value.require_valid.return_value.first.return_value = Mock(
            spec=Token
        )
        self.user = AuthUtils.add_permission_to_user_by_name(
            "freight.setup_contract_handler", self.user
        )
        handler = ContractHandler.objects.create(
            organization=self.alliance,
            character=self.main_ownership,
            operation_mode=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
        )

        handler.update_contracts_esi()

        # should only contain the right contracts
        contract_ids = [
            x["contract_id"]
            for x in Contract.objects.filter(
                status__exact=Contract.Status.OUTSTANDING
            ).values("contract_id")
        ]
        self.assertCountEqual(
            contract_ids, [149409005, 149409014, 149409006, 149409015]
        )

        # 2nd run should not update anything, but reset last_sync
        Contract.objects.all().delete()
        handler.last_sync = None
        handler.save()
        handler.update_contracts_esi()
        self.assertEqual(Contract.objects.count(), 0)
        handler.refresh_from_db()
        self.assertIsNotNone(handler.last_sync)

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_MY_CORPORATION)
    @patch(MODULE_PATH + ".Token")
    @patch(MODULE_PATH + ".esi")
    def test_sync_contracts_for_my_corporation(self, mock_esi, mock_Token):
        mock_Contracts = mock_esi.client.Contracts
        mock_Contracts.get_corporations_corporation_id_contracts.side_effect = (
            self.esi_get_corporations_corporation_id_contracts
        )
        mock_Token.objects.filter.return_value.require_scopes.return_value.require_valid.return_value.first.return_value = Mock(
            spec=Token
        )
        self.user = AuthUtils.add_permission_to_user_by_name(
            "freight.setup_contract_handler", self.user
        )
        handler = ContractHandler.objects.create(
            organization=self.corporation,
            character=self.main_ownership,
            operation_mode=FREIGHT_OPERATION_MODE_MY_CORPORATION,
        )

        # run manager sync
        handler.update_contracts_esi()
        handler.refresh_from_db()

        # should only contain the right contracts
        contract_ids = [
            x["contract_id"]
            for x in Contract.objects.filter(
                status__exact=Contract.Status.OUTSTANDING
            ).values("contract_id")
        ]
        self.assertCountEqual(
            contract_ids,
            [
                149409016,
                149409061,
                149409062,
                149409063,
                149409064,
            ],
        )

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE)
    @patch(MODULE_PATH + ".Token")
    @patch(MODULE_PATH + ".esi")
    def test_sync_contracts_for_corp_in_alliance(self, mock_esi, mock_Token):
        mock_Contracts = mock_esi.client.Contracts
        mock_Contracts.get_corporations_corporation_id_contracts.side_effect = (
            self.esi_get_corporations_corporation_id_contracts
        )
        mock_Token.objects.filter.return_value.require_scopes.return_value.require_valid.return_value.first.return_value = Mock(
            spec=Token
        )
        self.user = AuthUtils.add_permission_to_user_by_name(
            "freight.setup_contract_handler", self.user
        )
        handler = ContractHandler.objects.create(
            organization=self.corporation,
            character=self.main_ownership,
            operation_mode=FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE,
        )

        # run manager sync
        handler.update_contracts_esi()

        # should only contain the right contracts
        contract_ids = [
            x["contract_id"]
            for x in Contract.objects.filter(
                status__exact=Contract.Status.OUTSTANDING
            ).values("contract_id")
        ]
        self.assertCountEqual(
            contract_ids,
            [149409016, 149409017, 149409061, 149409062, 149409063, 149409064],
        )

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_CORP_PUBLIC)
    @patch(MODULE_PATH + ".Token")
    @patch(MODULE_PATH + ".esi")
    def test_can_sync_contracts_for_corp_public(self, mock_esi, mock_Token):
        # create mocks
        mock_Contracts = mock_esi.client.Contracts
        mock_Contracts.get_corporations_corporation_id_contracts.side_effect = (
            self.esi_get_corporations_corporation_id_contracts
        )
        mock_Token.objects.filter.return_value.require_scopes.return_value.require_valid.return_value.first.return_value = Mock(
            spec=Token
        )
        self.user = AuthUtils.add_permission_to_user_by_name(
            "freight.setup_contract_handler", self.user
        )
        handler = ContractHandler.objects.create(
            organization=self.corporation,
            character=self.main_ownership,
            operation_mode=FREIGHT_OPERATION_MODE_CORP_PUBLIC,
        )

        # run manager sync
        handler.update_contracts_esi()

        # should only contain the right contracts
        contract_ids = [
            x["contract_id"]
            for x in Contract.objects.filter(
                status__exact=Contract.Status.OUTSTANDING
            ).values("contract_id")
        ]
        self.assertCountEqual(
            contract_ids,
            [
                149409016,
                149409061,
                149409062,
                149409063,
                149409064,
                149409017,
                149409018,
            ],
        )

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    @patch(MODULE_PATH + ".esi")
    @patch(MODULE_PATH + ".ContractHandler.token")
    def test_should_abort_on_general_exception(self, mock_token, mock_esi):
        # given
        mock_esi.client.Contracts.get_corporations_corporation_id_contracts.side_effect = (
            RuntimeError
        )
        self.user = AuthUtils.add_permission_to_user_by_name(
            "freight.setup_contract_handler", self.user
        )
        handler = ContractHandler.objects.create(
            organization=self.alliance,
            character=self.main_ownership,
            operation_mode=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
        )
        # when/then
        with self.assertRaises(RuntimeError):
            handler.update_contracts_esi()

    @patch(PATCH_FREIGHT_OPERATION_MODE, FREIGHT_OPERATION_MODE_MY_ALLIANCE)
    def test_operation_mode_friendly(self):
        handler = ContractHandler.objects.create(
            organization=self.alliance,
            operation_mode=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
            character=self.main_ownership,
        )
        self.assertEqual(handler.operation_mode_friendly, FREIGHT_OPERATION_MODES[0][1])

        handler.operation_mode = FREIGHT_OPERATION_MODE_MY_CORPORATION
        self.assertEqual(handler.operation_mode_friendly, FREIGHT_OPERATION_MODES[1][1])

        handler.operation_mode = FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE
        self.assertEqual(handler.operation_mode_friendly, FREIGHT_OPERATION_MODES[2][1])

        handler.operation_mode = FREIGHT_OPERATION_MODE_CORP_PUBLIC
        self.assertEqual(handler.operation_mode_friendly, FREIGHT_OPERATION_MODES[3][1])


class TestEveEntity(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        create_entities_from_characters()
        cls.alliance = EveEntity.objects.get(id=93000001)
        cls.corporation = EveEntity.objects.get(id=92000001)
        cls.character = EveEntity.objects.get(id=90000001)

    def test_str(self):
        self.assertEqual(str(self.character), "Bruce Wayne")

    def test_repr(self):
        expected = (
            f"EveEntity(id={self.character.id}, category='character', "
            "name='Bruce Wayne')"
        )
        self.assertEqual(repr(self.character), expected)

    def test_is_alliance(self):
        self.assertFalse(self.character.is_alliance)
        self.assertFalse(self.corporation.is_alliance)
        self.assertTrue(self.alliance.is_alliance)

    def test_is_corporation(self):
        self.assertFalse(self.character.is_corporation)
        self.assertTrue(self.corporation.is_corporation)
        self.assertFalse(self.alliance.is_corporation)

    def test_is_character(self):
        self.assertTrue(self.character.is_character)
        self.assertFalse(self.corporation.is_character)
        self.assertFalse(self.alliance.is_character)

    def test_avatar_url_alliance(self):
        expected = "https://images.evetech.net/alliances/93000001/logo?size=128"
        self.assertEqual(self.alliance.icon_url(), expected)

    def test_avatar_url_corporation(self):
        expected = "https://images.evetech.net/corporations/92000001/logo?size=128"
        self.assertEqual(self.corporation.icon_url(), expected)

    def test_avatar_url_character(self):
        expected = "https://images.evetech.net/characters/90000001/portrait?size=128"
        self.assertEqual(self.character.icon_url(), expected)


class TestFreight(NoSocketsTestCase):
    def test_get_category_for_operation_mode_1(self):
        self.assertEqual(
            Freight.category_for_operation_mode(FREIGHT_OPERATION_MODE_MY_ALLIANCE),
            EveEntity.CATEGORY_ALLIANCE,
        )
        self.assertEqual(
            Freight.category_for_operation_mode(FREIGHT_OPERATION_MODE_MY_CORPORATION),
            EveEntity.CATEGORY_CORPORATION,
        )
        self.assertEqual(
            Freight.category_for_operation_mode(
                FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE
            ),
            EveEntity.CATEGORY_CORPORATION,
        )
        self.assertEqual(
            Freight.category_for_operation_mode(FREIGHT_OPERATION_MODE_CORP_PUBLIC),
            EveEntity.CATEGORY_CORPORATION,
        )



"""Pricing and location models."""

from typing import List, Optional

from django.core.exceptions import ValidationError
from django.core.validators import MinValueValidator
from django.db import models
from django.utils.translation import gettext_lazy as _

from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from freight import __title__
from freight.app_settings import FREIGHT_FULL_ROUTE_NAMES
from freight.managers import LocationManager, PricingManager

from .contract_handlers import ContractHandler

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Location(models.Model):
    """An Eve Online courier contract location: station or Upwell structure"""

    class Category(models.IntegerChoices):
        """A location category."""

        STATION_ID = 3, "station"
        STRUCTURE_ID = 65, "structure"
        UNKNOWN_ID = 0, "(unknown)"

    id = models.BigIntegerField(
        primary_key=True,
        validators=[MinValueValidator(0)],
        verbose_name=_("id"),
        help_text="Eve Online location ID, "
        "either item ID for stations or structure ID for structures",
    )

    category_id = models.PositiveIntegerField(
        choices=Category.choices,
        default=Category.UNKNOWN_ID,
        help_text="Eve Online category ID",
    )
    name = models.CharField(
        max_length=100,
        db_index=True,
        verbose_name=_("name"),
        help_text="In-game name of this station or structure",
    )
    solar_system_id = models.PositiveIntegerField(
        default=None, null=True, blank=True, help_text="Eve Online solar system ID"
    )
    type_id = models.PositiveIntegerField(
        default=None, null=True, blank=True, help_text="Eve Online type ID"
    )

    objects = LocationManager()

    class Meta:
        verbose_name = _("location")
        verbose_name_plural = _("locations")

    def __str__(self):
        return self.name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(pk={self.pk}, name='{self.name}')"

    @property
    def category(self):
        """Return category ID for this location."""
        return self.category_id

    @property
    def solar_system_name(self):
        """Return solar system name for this location.."""
        return self.name.split(" ", 1)[0]

    @property
    def location_name(self):
        """Return name of this location."""
        return self.name.rsplit("-", 1)[1].strip()

    @classmethod
    def get_esi_scopes(cls):
        """Return ESI scopes required to fetch this data."""
        return ["esi-universe.read_structures.v1"]


class Pricing(models.Model):
    """Pricing for a courier route"""

    start_location = models.ForeignKey(
        Location,
        on_delete=models.CASCADE,
        verbose_name=_("start location"),
        related_name="+",
        help_text=_("Starting station or structure for courier route"),
    )
    end_location = models.ForeignKey(
        Location,
        on_delete=models.CASCADE,
        verbose_name=_("end location"),
        related_name="+",
        help_text=_("Destination station or structure for courier route"),
    )

    collateral_min = models.BigIntegerField(
        default=None,
        null=True,
        blank=True,
        validators=[MinValueValidator(0)],
        verbose_name=_("collateral minimum"),
        help_text=_("Minimum required collateral in ISK"),
    )
    collateral_max = models.BigIntegerField(
        default=None,
        null=True,
        blank=True,
        validators=[MinValueValidator(0)],
        verbose_name=_("collateral maximum"),
        help_text=_("Maximum allowed collateral in ISK"),
    )
    days_to_expire = models.PositiveIntegerField(
        default=None,
        null=True,
        blank=True,
        validators=[MinValueValidator(1)],
        verbose_name=_("days to expire"),
        help_text=_("Recommended days for contracts to expire"),
    )
    days_to_complete = models.PositiveIntegerField(
        default=None,
        null=True,
        blank=True,
        validators=[MinValueValidator(1)],
        verbose_name=_("days to complete"),
        help_text=_("Recommended days for contract completion"),
    )
    details = models.TextField(
        default=None,
        null=True,
        blank=True,
        verbose_name=_("details"),
        help_text=_("Text with additional instructions for using this pricing"),
    )
    is_active = models.BooleanField(
        default=True,
        verbose_name=_("is active"),
        help_text=_("Disabled pricings will not be used or shown"),
    )
    is_default = models.BooleanField(
        default=False,
        verbose_name=_("is default"),
        help_text=_(
            "The default pricing will be preselected in the calculator. "
            "Please make sure to only mark one pricing as default."
        ),
    )
    is_bidirectional = models.BooleanField(
        default=True,
        verbose_name=_("is bidirectional"),
        help_text=_(
            "Whether this pricing is valid for contracts "
            "in either direction or only the one specified"
        ),
    )
    price_base = models.FloatField(
        default=None,
        null=True,
        blank=True,
        validators=[MinValueValidator(0)],
        verbose_name=_("price base"),
        help_text=_("Base price in ISK"),
    )
    price_min = models.FloatField(
        default=None,
        null=True,
        blank=True,
        validators=[MinValueValidator(0)],
        verbose_name=_("price minimum"),
        help_text=_("Minimum total price in ISK"),
    )
    price_per_volume = models.FloatField(
        default=None,
        null=True,
        blank=True,
        validators=[MinValueValidator(0)],
        verbose_name=_("price markup from volume"),
        help_text=_("Add-on price per m3 volume in ISK"),
    )
    price_per_collateral_percent = models.FloatField(
        default=None,
        null=True,
        blank=True,
        validators=[MinValueValidator(0)],
        verbose_name=_("price markup from collateral"),
        help_text=_("Add-on price in percent of collateral"),
    )
    volume_max = models.FloatField(
        default=None,
        null=True,
        blank=True,
        validators=[MinValueValidator(0)],
        verbose_name=_("volume maximum"),
        help_text=_("Maximum allowed volume in m3"),
    )
    volume_min = models.FloatField(
        default=None,
        null=True,
        blank=True,
        validators=[MinValueValidator(0)],
        verbose_name=_("volume minimum"),
        help_text=_("Minimum allowed volume in m3"),
    )
    use_price_per_volume_modifier = models.BooleanField(
        default=False,
        verbose_name=_("price per volume modifier enabled"),
        help_text=_("Whether the global price per volume modifier is used"),
    )

    objects = PricingManager()

    class Meta:
        unique_together = (("start_location", "end_location"),)
        verbose_name = _("pricing")
        verbose_name_plural = _("pricings")

    def save(self, *args, **kwargs) -> None:
        update_contracts = kwargs.pop("update_contracts", True)
        super().save(*args, **kwargs)
        if update_contracts:
            self._update_contracts()

    def _update_contracts(self):
        from freight.tasks import update_contracts_pricing

        update_contracts_pricing.delay()

    def __str__(self) -> str:
        return self.name

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(pk={self.pk}, name='{self.name_full}')"

    def clean(self):
        if (
            self.price_base is None
            and self.price_min is None
            and self.price_per_volume is None
            and self.price_per_collateral_percent is None
        ):
            raise ValidationError(_("You must specify at least one price component"))

        if self.start_location_id and self.end_location_id:
            if (
                Pricing.objects.filter(
                    start_location=self.end_location,
                    end_location=self.start_location,
                    is_bidirectional=True,
                ).exists()
                and self.is_bidirectional
            ):
                raise ValidationError(
                    _(
                        "There already exists a bidirectional pricing for this route. "
                        "Please set this pricing to non-bidirectional to save it. "
                        "And after you must also set the other pricing to "
                        "non-bidirectional."
                    )
                )

            if (
                Pricing.objects.filter(
                    start_location=self.end_location,
                    end_location=self.start_location,
                    is_bidirectional=False,
                ).exists()
                and self.is_bidirectional
            ):
                raise ValidationError(
                    _(
                        "There already exists a non bidirectional pricing for "
                        "this route. You need to mark this pricing as "
                        "non-bidirectional too to continue."
                    )
                )

    @property
    def name(self) -> str:
        """Return default pricing name."""
        return self._name(FREIGHT_FULL_ROUTE_NAMES)

    @property
    def name_full(self) -> str:
        """Return full pricing name."""
        return self._name(full_name=True)

    @property
    def name_short(self) -> str:
        """Return short pricing name."""
        return self._name(full_name=False)

    def _name(self, full_name: bool) -> str:
        if full_name:
            start_name = self.start_location.name
            end_name = self.end_location.name
        else:
            start_name = self.start_location.solar_system_name
            end_name = self.end_location.solar_system_name

        arrow = "<->" if self.is_bidirectional else "->"
        route_name = f"{start_name} {arrow} {end_name}"
        return route_name

    def price_per_volume_modifier(self):
        """returns the effective price per volume modifier or None"""
        if not self.use_price_per_volume_modifier:
            modifier = None

        else:
            handler = ContractHandler.objects.first()
            if handler:
                modifier = handler.price_per_volume_modifier

            else:
                modifier = None

        return modifier

    def price_per_volume_eff(self):
        """ "returns price per volume incl. potential modifier or None"""
        if not self.price_per_volume:
            price_per_volume = None
        else:
            price_per_volume = self.price_per_volume
            modifier = self.price_per_volume_modifier()
            if modifier:
                price_per_volume = max(
                    0, price_per_volume + (price_per_volume * modifier / 100)
                )

        return price_per_volume

    def requires_volume(self) -> bool:
        """whether this pricing required volume to be specified"""
        return (self.price_per_volume is not None and self.price_per_volume != 0) or (
            self.volume_min is not None and self.volume_min != 0
        )

    def requires_collateral(self) -> bool:
        """whether this pricing required collateral to be specified"""
        return (
            self.price_per_collateral_percent is not None
            and self.price_per_collateral_percent != 0
        ) or (self.collateral_min is not None and self.collateral_min != 0)

    def is_fix_price(self) -> bool:
        """whether this pricing is a fix price"""
        return (
            self.price_base is not None
            and self.price_min is None
            and self.price_per_volume is None
            and self.price_per_collateral_percent is None
        )

    def get_calculated_price(self, volume: float, collateral: float) -> float:
        """returns the calculated price for the given parameters"""

        if not volume:
            volume = 0

        if not collateral:
            collateral = 0

        if volume < 0:
            raise ValueError("volume can not be negative")
        if collateral < 0:
            raise ValueError("collateral can not be negative")

        volume = float(volume)
        collateral = float(collateral)

        price_base = 0 if not self.price_base else self.price_base
        price_min = 0 if not self.price_min else self.price_min

        price_per_volume_eff = self.price_per_volume_eff()
        if not price_per_volume_eff:
            price_per_volume = 0
        else:
            price_per_volume = price_per_volume_eff

        price_per_collateral_percent = (
            0
            if not self.price_per_collateral_percent
            else self.price_per_collateral_percent
        )

        return max(
            price_min,
            (
                price_base
                + volume * price_per_volume
                + collateral * (price_per_collateral_percent / 100)
            ),
        )

    def get_contract_price_check_issues(
        self, volume: float, collateral: float, reward: float = None
    ) -> Optional[List[str]]:
        """Return list of validation error messages or none if ok."""
        if volume and volume < 0:
            raise ValueError("volume can not be negative")
        if collateral and collateral < 0:
            raise ValueError("collateral can not be negative")
        if reward and reward < 0:
            raise ValueError("reward can not be negative")

        issues = []
        if volume is not None and self.volume_min and volume < self.volume_min:
            issues.append(
                f"below the minimum required volume of {self.volume_min:,.0f} m3"
            )
        if volume is not None and self.volume_max and volume > self.volume_max:
            issues.append(
                f"exceeds the maximum allowed volume of {self.volume_max:,.0f} m3"
            )
        if (
            collateral is not None
            and self.collateral_max
            and collateral > self.collateral_max
        ):
            issues.append(
                "exceeds the maximum allowed collateral "
                f"of {self.collateral_max:,.0f} ISK"
            )
        if (
            collateral is not None
            and self.collateral_min
            and collateral < self.collateral_min
        ):
            issues.append(
                "below the minimum required collateral "
                f"of {self.collateral_min:,.0f} ISK"
            )
        if reward is not None:
            calculated_price = self.get_calculated_price(volume, collateral)
            if reward < calculated_price:
                issues.append(
                    "reward is below the calculated price "
                    f"of {calculated_price:,.0f} ISK"
                )
        if len(issues) == 0:
            return None
        return issues



"""Contract Handler models."""

import hashlib
import json
from datetime import timedelta
from typing import List

from django.conf import settings
from django.core.serializers.json import DjangoJSONEncoder
from django.core.validators import MinValueValidator
from django.db import models, transaction
from django.utils.timezone import now
from django.utils.translation import gettext_lazy as _
from esi.errors import TokenError
from esi.models import Token

from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from freight import __title__
from freight.app_settings import (
    FREIGHT_CONTRACT_SYNC_GRACE_MINUTES,
    FREIGHT_OPERATION_MODE,
    FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE,
    FREIGHT_OPERATION_MODE_CORP_PUBLIC,
    FREIGHT_OPERATION_MODE_MY_ALLIANCE,
    FREIGHT_OPERATION_MODE_MY_CORPORATION,
    FREIGHT_OPERATION_MODES,
)
from freight.constants import AVATAR_SIZE
from freight.helpers import get_or_create_eve_character
from freight.managers import EveEntityManager
from freight.providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class Freight(models.Model):
    """Meta model for global app permissions"""

    class Meta:
        managed = False
        default_permissions = ()
        permissions = (
            ("add_location", "Can add / update locations"),
            ("basic_access", "Can access this app"),
            ("setup_contract_handler", "Can setup contract handler"),
            ("use_calculator", "Can use the calculator"),
            ("view_contracts", "Can view the contracts list"),
            ("view_statistics", "Can view freight statistics"),
        )

    @classmethod
    def operation_mode_friendly(cls, operation_mode) -> str:
        """returns user friendly description of operation mode"""
        msg = [(x, y) for x, y in FREIGHT_OPERATION_MODES if x == operation_mode]
        if len(msg) != 1:
            raise ValueError("Undefined mode")
        return msg[0][1]

    @staticmethod
    def category_for_operation_mode(mode: str) -> str:
        """Eve Entity category for given operation mode."""
        if mode == FREIGHT_OPERATION_MODE_MY_ALLIANCE:
            return EveEntity.CATEGORY_ALLIANCE
        return EveEntity.CATEGORY_CORPORATION


class EveEntity(models.Model):
    """An Eve entity like a corporation or a character"""

    CATEGORY_ALLIANCE = "alliance"
    CATEGORY_CHARACTER = "character"
    CATEGORY_CORPORATION = "corporation"

    CATEGORY_CHOICES = (
        (CATEGORY_ALLIANCE, "Alliance"),
        (CATEGORY_CORPORATION, "Corporation"),
        (CATEGORY_CHARACTER, "Character"),
    )

    id = models.IntegerField(
        primary_key=True, validators=[MinValueValidator(0)], verbose_name=_("id")
    )
    category = models.CharField(
        max_length=32, choices=CATEGORY_CHOICES, verbose_name=_("category")
    )
    name = models.CharField(max_length=254, verbose_name=_("name"))

    objects = EveEntityManager()

    def __str__(self):
        return self.name

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(id={self.id}, category='{self.category}', "
            f"name='{self.name}')"
        )

    @property
    def is_alliance(self) -> bool:
        """Return True if entity is an alliance, else False."""
        return self.category == self.CATEGORY_ALLIANCE

    @property
    def is_corporation(self) -> bool:
        """Return True if entity is an corporation, else False."""
        return self.category == self.CATEGORY_CORPORATION

    @property
    def is_character(self) -> bool:
        """Return True if entity is a character, else False."""
        return self.category == self.CATEGORY_CHARACTER

    def icon_url(self, size=AVATAR_SIZE) -> str:
        """Url to an icon image for this organization."""
        if self.category == self.CATEGORY_ALLIANCE:
            return EveAllianceInfo.generic_logo_url(self.id, size=size)

        if self.category == self.CATEGORY_CORPORATION:
            return EveCorporationInfo.generic_logo_url(self.id, size=size)

        if self.category == self.CATEGORY_CHARACTER:
            return EveCharacter.generic_portrait_url(self.id, size=size)

        raise NotImplementedError(
            f"Avatar URL not implemented for category {self.category}"
        )


class ContractHandler(models.Model):
    """Handler for syncing of contracts belonging to an alliance or corporation."""

    # errors
    ERROR_NONE = 0
    ERROR_TOKEN_INVALID = 1
    ERROR_TOKEN_EXPIRED = 2
    ERROR_INSUFFICIENT_PERMISSIONS = 3
    ERROR_NO_CHARACTER = 4
    ERROR_ESI_UNAVAILABLE = 5
    ERROR_OPERATION_MODE_MISMATCH = 6
    ERROR_UNKNOWN = 99

    ERRORS_LIST = [
        (ERROR_NONE, "No error"),
        (ERROR_TOKEN_INVALID, "Invalid token"),
        (ERROR_TOKEN_EXPIRED, "Expired token"),
        (ERROR_INSUFFICIENT_PERMISSIONS, "Insufficient permissions"),
        (ERROR_NO_CHARACTER, "No character set for fetching alliance contacts"),
        (ERROR_ESI_UNAVAILABLE, "ESI API is currently unavailable"),
        (
            ERROR_OPERATION_MODE_MISMATCH,
            "Operation mode does not match with current setting",
        ),
        (ERROR_UNKNOWN, "Unknown error"),
    ]

    organization = models.OneToOneField(
        EveEntity,
        on_delete=models.CASCADE,
        primary_key=True,
        verbose_name=_("organization"),
    )
    character = models.ForeignKey(
        CharacterOwnership,
        on_delete=models.SET_DEFAULT,
        default=None,
        null=True,
        verbose_name=_("character"),
        related_name="+",
        help_text=_("Character used for syncing contracts"),
    )
    operation_mode = models.CharField(
        max_length=32,
        default=FREIGHT_OPERATION_MODE_MY_ALLIANCE,
        verbose_name=_("operation mode"),
        help_text=_("Defines what kind of contracts are synced"),
    )
    price_per_volume_modifier = models.FloatField(
        default=None,
        null=True,
        blank=True,
        verbose_name=_("price per volume modifier"),
        help_text=_(
            "Global modifier for price per volume in percent, e.g. 2.5 = +2.5%"
        ),
    )
    version_hash = models.CharField(
        max_length=32,
        null=True,
        default=None,
        blank=True,
        verbose_name=_("version hash"),
        help_text=_("hash to identify changes to contracts"),
    )
    last_sync = models.DateTimeField(
        null=True,
        default=None,
        blank=True,
        verbose_name=_("last sync at"),
        help_text=_("when the last sync happened"),
    )
    last_error = models.IntegerField(
        choices=ERRORS_LIST,
        default=ERROR_NONE,
        help_text="error that occurred at the last sync attempt (if any)",
    )  # TODO: Remove with next migration - no longer used

    class Meta:
        verbose_name = _("contract handler")
        verbose_name_plural = _("contract handler")

    def __str__(self):
        return str(self.organization.name)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(pk={self.pk}, "
            f"organization='{self.organization.name}')"
        )

    @property
    def operation_mode_friendly(self) -> str:
        """returns user friendly description of operation mode"""
        return Freight.operation_mode_friendly(self.operation_mode)

    # @property
    # def last_error_message_friendly(self) -> str:
    #     msg = [(x, y) for x, y in self.ERRORS_LIST if x == self.last_error]
    #     return msg[0][1] if len(msg) > 0 else "Undefined error"

    @classmethod
    def get_esi_scopes(cls) -> List[str]:
        """Return list of required ESI scopes to fetch contracts."""
        return [
            "esi-contracts.read_corporation_contracts.v1",
            "esi-universe.read_structures.v1",
        ]

    @property
    def is_sync_ok(self) -> bool:
        """returns true if they have been no errors
        and last syncing occurred within alloted time
        """
        grace_deadline = now() - timedelta(minutes=FREIGHT_CONTRACT_SYNC_GRACE_MINUTES)
        return self.last_sync and self.last_sync > grace_deadline

    def get_availability_text_for_contracts(self) -> str:
        """returns a text detailing the availability choice for this setup"""

        if self.operation_mode == FREIGHT_OPERATION_MODE_MY_ALLIANCE:
            extra_text = "[My Alliance]"

        elif self.operation_mode == FREIGHT_OPERATION_MODE_MY_CORPORATION:
            extra_text = "[My Corporation]"

        else:
            extra_text = ""

        return f"Private ({self.organization.name}) {extra_text}"

    def token(self) -> Token:
        """Returns an valid esi token for the contract handler.

        Raises exception on any error
        """
        token = (
            Token.objects.filter(
                user=self.character.user,
                character_id=self.character.character.character_id,
            )
            .require_scopes(self.get_esi_scopes())
            .require_valid()
            .first()
        )
        if not token:
            raise TokenError(f"{self}: No valid token found")

        return token

    def update_contracts_esi(self, force_sync=False) -> bool:
        """Update contracts from ESI."""
        self._validate_update_readiness()
        token = self.token()
        contracts = esi.client.Contracts.get_corporations_corporation_id_contracts(
            token=token.valid_access_token(),
            corporation_id=self.character.character.corporation_id,
        ).results()
        if settings.DEBUG:
            self._save_contract_to_file(contracts)

        self._process_contracts_from_esi(contracts, token, force_sync)
        self.last_sync = now()
        self.save(update_fields=["last_sync"])

    def _validate_update_readiness(self):
        if self.operation_mode != FREIGHT_OPERATION_MODE:
            raise ValueError(f"{self}: Current operation mode not matching the handler")

        if self.character is None:
            raise ValueError(f"{self}: No character configured to sync")

        if not self.character.user.has_perm("freight.setup_contract_handler"):
            raise ValueError(
                f"{self}: Character does not have sufficient permission to sync"
            )

    def _save_contract_to_file(self, contracts):
        """saves raw contracts to file for debugging"""
        with open("contracts_raw.json", "w", encoding="utf-8") as file:
            json.dump(contracts, file, cls=DjangoJSONEncoder, sort_keys=True, indent=4)

    def _process_contracts_from_esi(
        self, contracts_all: list, token: object, force_sync: bool
    ):
        # 1st filter: reduce to courier contracts assigned to handler org
        contracts_courier = [
            x
            for x in contracts_all
            if x["type"] == "courier"
            and int(x["assignee_id"]) == int(self.organization.id)
        ]

        # 2nd filter: remove contracts not in scope due to operation mode
        contracts = []
        for contract in contracts_courier:
            issuer, _ = get_or_create_eve_character(contract["issuer_id"])

            assignee_id = int(contract["assignee_id"])
            issuer_corporation_id = int(issuer.corporation_id)
            issuer_alliance_id = int(issuer.alliance_id) if issuer.alliance_id else None

            if self.operation_mode == FREIGHT_OPERATION_MODE_MY_ALLIANCE:
                in_scope = issuer_alliance_id == assignee_id

            elif self.operation_mode == FREIGHT_OPERATION_MODE_MY_CORPORATION:
                in_scope = assignee_id == issuer_corporation_id

            elif self.operation_mode == FREIGHT_OPERATION_MODE_CORP_IN_ALLIANCE:
                in_scope = issuer_alliance_id == int(
                    self.character.character.alliance_id
                )

            elif self.operation_mode == FREIGHT_OPERATION_MODE_CORP_PUBLIC:
                in_scope = True

            else:
                raise NotImplementedError(
                    f"Unsupported operation mode: {self.operation_mode}"
                )
            if in_scope:
                contracts.append(contract)

        # determine if contracts have changed by comparing their hashes
        new_version_hash = hashlib.md5(
            json.dumps(contracts, cls=DjangoJSONEncoder).encode("utf-8")
        ).hexdigest()
        if force_sync or new_version_hash != self.version_hash:
            self._store_contract_from_esi(contracts, new_version_hash, token)
        else:
            logger.info("%s: Contracts are unchanged.", self)

    # pylint: disable = no-member
    def _store_contract_from_esi(
        self, contracts: list, new_version_hash: str, token: Token
    ) -> None:
        logger.info("%s: Storing update with %d contracts", self, len(contracts))
        with transaction.atomic():
            self.version_hash = new_version_hash
            for contract in contracts:
                try:
                    self.contracts.update_or_create_from_dict(
                        handler=self, contract=contract, token=token
                    )
                except OSError:
                    logger.exception(
                        "%s: An unexpected error ocurred while trying to load contract "
                        "%s",
                        self,
                        (
                            contract["contract_id"]
                            if "contract_id" in contract
                            else "Unknown"
                        ),
                    )

        self.contracts.update_pricing()



"""Admin site for Freight."""

# pylint: disable=missing-class-docstring, missing-function-docstring


from django.conf import settings
from django.contrib import admin
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

from . import tasks
from .models import (
    Contract,
    ContractCustomerNotification,
    ContractHandler,
    EveEntity,
    Location,
    Pricing,
)
from .tasks import update_locations


@admin.register(Location)
class LocationAdmin(admin.ModelAdmin):
    list_display = ("id", "name", "_category", "_solar_system")
    list_filter = ("category_id",)
    search_fields = ["name"]
    list_select_related = True
    list_display_links = None
    actions = ["update_location"]

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    @admin.display(ordering="category_id", description=_("category"))
    def _category(self, obj):
        return obj.get_category_id_display()

    @admin.display(description=_("solar system"))
    def _solar_system(self, obj):
        return obj.solar_system_name

    @admin.action(description=_("Update selected locations from ESI"))
    def update_location(self, request, queryset):
        location_ids = []
        for obj in queryset:
            location_ids.append(obj.pk)

        update_locations.delay(location_ids)
        self.message_user(
            request,
            (
                _(
                    "Started updating %d locations. "
                    "This can take a short while to complete."
                )
                % len(location_ids)
            ),
        )


@admin.register(Pricing)
class PricingAdmin(admin.ModelAdmin):
    list_display = (
        "_name",
        "start_location",
        "end_location",
        "_bidirectional",
        "_default",
        "_active",
    )
    list_filter = (
        "is_bidirectional",
        "is_active",
        ("start_location", admin.RelatedOnlyFieldListFilter),
        ("end_location", admin.RelatedOnlyFieldListFilter),
    )
    list_select_related = True

    @admin.display(boolean=True, description=_("bidirectional"))
    def _bidirectional(self, obj):
        return obj.is_bidirectional

    @admin.display(description=_("name"))
    def _name(self, obj):
        return format_html('<span style="white-space: nowrap;">{}</span>', obj.name)

    @admin.display(boolean=True, description=_("active"))
    def _active(self, obj):
        return obj.is_active

    @admin.display(boolean=True, description=_("default"))
    def _default(self, obj):
        return obj.is_default


@admin.register(ContractHandler)
class ContractHandlerAdmin(admin.ModelAdmin):
    list_display = (
        "organization",
        "character",
        "operation_mode",
        "last_sync",
        "_is_sync_ok",
    )
    actions = ("start_sync", "send_notifications", "update_pricing")
    readonly_fields = (
        "organization",
        "character",
        "operation_mode",
        "version_hash",
        "last_sync",
    )

    def get_fields(self, request, obj=None):
        """Do not show some specific fields."""
        fields = super().get_fields(request, obj)
        return [name for name in fields if name != "last_error"]

    def has_add_permission(self, *args, **kwargs):
        return False

    @admin.display(boolean=True, description=_("sync ok"))
    def _is_sync_ok(self, obj):
        return obj.is_sync_ok

    @admin.action(description=_("Fetch contracts from Eve Online server"))
    def start_sync(self, request, queryset):
        for obj in queryset:
            tasks.run_contracts_sync.delay(force_sync=True)
            self.message_user(request, _("Started syncing contracts for: %s") % obj)

    @admin.action(description=_("Send notifications for outstanding contracts"))
    def send_notifications(self, request, queryset):
        for obj in queryset:
            tasks.send_contract_notifications.delay(force_sent=True)
            self.message_user(
                request, _("Started sending notifications for: %s ") % obj
            )

    @admin.action(description=_("Update pricing info for all contracts"))
    def update_pricing(self, request, queryset):
        del queryset
        tasks.update_contracts_pricing.delay()
        self.message_user(
            request, "Started updating pricing relations for all contracts"
        )


@admin.register(Contract)
class ContractAdmin(admin.ModelAdmin):
    list_display = [
        "contract_id",
        "status",
        "date_issued",
        "issuer",
        "pricing",
        "_pilots_notified",
        "_customer_notified",
    ]
    list_filter = (
        "status",
        ("issuer", admin.RelatedOnlyFieldListFilter),
        ("pricing", admin.RelatedOnlyFieldListFilter),
    )
    search_fields = ["issuer"]
    list_select_related = (
        "issuer",
        "pricing",
        "pricing__start_location",
        "pricing__end_location",
    )
    actions = ["send_pilots_notification", "send_customer_notification"]

    def get_queryset(self, request):
        qs = super().get_queryset(request)
        return qs.prefetch_related("customer_notifications")

    def has_add_permission(self, *args, **kwargs):
        return False

    def has_change_permission(self, *args, **kwargs):
        return False

    @admin.display(boolean=True, description=_("pilots notified"))
    def _pilots_notified(self, contract):
        if contract.pricing_id is None:
            return None
        return contract.date_notified is not None

    @admin.display(description=_("customer notified"))
    def _customer_notified(self, contract):
        if contract.pricing_id is None:
            return "?"
        notifications = [x.status for x in contract.customer_notifications.all()]
        if not notifications:
            return None
        return ", ".join(sorted(notifications, reverse=True))

    @admin.action(
        description=_("Sent pilots notification for selected contracts to Discord")
    )
    def send_pilots_notification(self, request, queryset):
        for obj in queryset:
            obj.send_pilot_notification()
            self.message_user(
                request,
                _("Sent pilots notification for contract %d to Discord")
                % obj.contract_id,
            )

    @admin.action(
        description=_("Sent customer notification for selected contracts to Discord")
    )
    def send_customer_notification(self, request, queryset):
        for obj in queryset:
            obj.send_customer_notification(force_sent=True)
            self.message_user(
                request,
                _("Sent customer notification for contract %d to Discord")
                % obj.contract_id,
            )


if settings.DEBUG:

    @admin.register(ContractCustomerNotification)
    class ContractCustomerNotificationAdmin(admin.ModelAdmin):
        def has_add_permission(self, *args, **kwargs):
            return False

        def has_change_permission(self, *args, **kwargs):
            return False

    @admin.register(EveEntity)
    class EveEntityAdmin(admin.ModelAdmin):
        list_display = ("name", "category")
        list_filter = ("category",)

        def has_add_permission(self, *args, **kwargs):
            return False

        def has_change_permission(self, *args, **kwargs):
            return False



import datetime as dt

from django.utils.timezone import now

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo

from ...models import Contract, ContractHandler, Pricing


def create_contract_handler(**kwargs):
    kwargs["organization_id"] = kwargs["character"].character.corporation_id
    return ContractHandler.objects.create(**kwargs)


def create_contract(**kwargs):
    contract_id = int(dt.datetime.now().timestamp())
    params = {
        "contract_id": contract_id,
        "collateral": 1_000_000_000,
        "date_expired": now() + dt.timedelta(days=3),
        "date_issued": now(),
        "days_to_complete": 3,
        "for_corporation": False,
        "reward": 100_000_000,
        "status": Contract.Status.OUTSTANDING,
        "title": f"Test contract #{contract_id}",
        "volume": 20_000,
    }
    params.update(kwargs)
    if "end_location" not in params and "end_location_id" not in params:
        params["end_location_id"] = 1022167642188
    if "issuer_corporation" not in params and "issuer_corporation_id" not in params:
        params["issuer_corporation"] = EveCorporationInfo.objects.get(
            corporation_id=92000001
        )
    if "issuer" not in params and "issuer_id" not in params:
        params["issuer"] = EveCharacter.objects.get(character_id=90000002)
    if "start_location" not in params and "start_location_id" not in params:
        params["start_location_id"] = 60003760
    return Contract.objects.create(**params)


def create_pricing(update_contracts: bool = False, **kwargs) -> Pricing:
    obj = Pricing(**kwargs)
    obj.save(update_contracts=update_contracts)
    return obj



"""scripts generates large amount of moons for load testing

This script can be executed directly from shell.
"""

import os
import sys
from pathlib import Path

myauth_dir = Path(__file__).parent.parent.parent.parent.parent / "myauth"
sys.path.insert(0, str(myauth_dir))

import django  # noqa: E402

# init and setup django project
os.environ.setdefault("DJANGO_SETTINGS_MODULE", "myauth.settings.local")
django.setup()

#######################
# main script
import datetime as dt  # noqa: E402
import random  # noqa: E402
from pathlib import Path  # noqa: E402

from django.utils.timezone import now  # noqa: E402

from allianceauth.eveonline.models import EveCharacter, EveCorporationInfo  # noqa: E402

from freight.models import Contract, ContractHandler, Location, Pricing  # noqa: E402

MAX_CONTRACTS = 10

print(f"Generating {MAX_CONTRACTS} contracts...")
jita_44, _ = Location.objects.get_or_create(
    id=60003760,
    defaults={
        "name": "Jita IV - Moon 4 - Caldari Navy Assembly Plant",
        "solar_system_id": 30000142,
        "type_id": 52678,
        "category_id": 3,
    },
)
enaluri_5, _ = Location.objects.get_or_create(
    id=60015068,
    defaults={
        "name": "Enaluri V - State Protectorate Assembly Plant",
        "solar_system_id": 30045339,
        "type_id": 1529,
        "category_id": 3,
    },
)
pricing, _ = Pricing.objects.get_or_create(
    start_location=jita_44,
    end_location=enaluri_5,
    defaults={
        "price_base": 200_000_000,
        "days_to_expire": 14,
        "days_to_complete": 3,
        "volume_max": 320_000,
        "details": "GENERATED PRICING FOR TESTING",
    },
)
handler = ContractHandler.objects.first()
acceptor = EveCharacter.objects.all().order_by("?").first()
acceptor_corporation = EveCorporationInfo.objects.get(
    corporation_id=acceptor.corporation_id
)
issuer = EveCharacter.objects.all().order_by("?").first()
issuer_corporation = EveCorporationInfo.objects.get(
    corporation_id=issuer.corporation_id
)
for _ in range(MAX_CONTRACTS):
    contract = Contract.objects.create(
        handler=handler,
        contract_id=random.randint(100_000_000, 200_000_000),
        acceptor=acceptor,
        acceptor_corporation=acceptor_corporation,
        collateral=random.randint(1_000_000_000, 5_000_000_000),
        date_accepted=now() - dt.timedelta(days=2),
        date_completed=now() - dt.timedelta(days=1),
        date_expired=now() + dt.timedelta(days=14),
        date_issued=now() - dt.timedelta(days=3),
        days_to_complete=7,
        end_location=enaluri_5,
        for_corporation=False,
        issuer_corporation=issuer_corporation,
        issuer=issuer,
        reward=random.randint(1_000_000, 500_000_000),
        start_location=jita_44,
        status=random.choice(
            [
                Contract.Status.OUTSTANDING,
                Contract.Status.OUTSTANDING,
                Contract.Status.IN_PROGRESS,
                Contract.Status.FINISHED,
                Contract.Status.FINISHED,
            ]
        ),
        title="GENERATED CONTRACT",
        volume=random.randint(10_000, 300_000),
        pricing=pricing,
    )
Contract.objects.update_pricing()
print("DONE")
